import dp_mobility_report.constants as const


def default_measure_selection() -> dict:
    return {
        const.DS_STATISTICS: const.SMAPE,
        const.MISSING_VALUES: const.SMAPE,
        const.TRIPS_OVER_TIME: const.JSD,
        const.TRIPS_PER_WEEKDAY: const.JSD,
        const.TRIPS_PER_HOUR: const.JSD,
        const.VISITS_PER_TILE: const.EMD,
        const.VISITS_PER_TILE_QUARTILES: const.SMAPE,
        const.VISITS_PER_TILE_OUTLIERS: const.SMAPE,
        const.VISITS_PER_TILE_RANKING: const.KT,
        const.VISITS_PER_TIME_TILE: const.EMD,
        const.OD_FLOWS: const.JSD,
        const.OD_FLOWS_RANKING: const.KT,
        const.OD_FLOWS_QUARTILES: const.SMAPE,
        const.TRAVEL_TIME: const.JSD,
        const.TRAVEL_TIME_QUARTILES: const.SMAPE,
        const.JUMP_LENGTH: const.JSD,
        const.JUMP_LENGTH_QUARTILES: const.SMAPE,
        const.TRIPS_PER_USER: const.EMD,  # TODO: implement JSD
        const.TRIPS_PER_USER_QUARTILES: const.SMAPE,
        const.USER_TIME_DELTA: const.JSD,
        const.USER_TIME_DELTA_QUARTILES: const.SMAPE,
        const.RADIUS_OF_GYRATION: const.JSD,
        const.RADIUS_OF_GYRATION_QUARTILES: const.SMAPE,
        const.USER_TILE_COUNT: const.JSD,
        const.USER_TILE_COUNT_QUARTILES: const.SMAPE,
        const.MOBILITY_ENTROPY: const.JSD,
        const.MOBILITY_ENTROPY_QUARTILES: const.SMAPE,
    }


def remove_excluded_analyses_from_measure_selection(
    measure_selection: dict, analysis_exclusion: list
) -> dict:
    def _is_included(x: str) -> bool:
        return not any([x.startswith(analysis) for analysis in analysis_exclusion])

    included_measures = {
        key: value for key, value in measure_selection.items() if _is_included(key)
    }
    return included_measures



from typing import List, Optional, Tuple, Type, Union

import numpy as np
import pandas as pd
from haversine import haversine
from pandas import DataFrame, Series

from dp_mobility_report.model.section import TupleSection
from dp_mobility_report.privacy import diff_privacy


def haversine_dist(coords: List[float]) -> float:
    # coords: provide coordinates as lat_start, lng_start, lat_end, lng_end
    return haversine((coords[0], coords[1]), (coords[2], coords[3]))


def cut_outliers(
    data: Series,
    min_value: Optional[Union[float, int]] = None,
    max_value: Optional[Union[float, int]] = None,
) -> Tuple:
    if (min_value is not None) and (max_value is not None) and (min_value > max_value):
        raise ValueError("min_value cannot be larger than max_value")
    if min_value is not None:
        data = data[data >= min_value]
    if max_value is not None:
        data = data[data <= max_value]
    return data


def hist_section(
    series: Union[np.ndarray, Series],
    eps: Optional[float],
    sensitivity: int,
    hist_max: Optional[Union[float, int]] = None,
    hist_min: Optional[Union[float, int]] = None,
    bin_range: Optional[Union[float, int]] = None,
    bin_type: Type = float,
    evalu: bool = False,
) -> TupleSection:
    epsi = get_epsi(evalu, eps, 6)
    epsi_quant = epsi * 5 if epsi is not None else None

    series = Series(series) if isinstance(series, np.ndarray) else series
    quartiles, moe_expmech = diff_privacy.quartiles_dp(series, epsi_quant, sensitivity)
    series = cut_outliers(
        series, min_value=quartiles["min"], max_value=quartiles["max"]
    )

    # determine bins for histogram
    # hist_min and hist_max determine how the output histogram looks like

    # max value of histogram: either given as input or determined by dp max
    hist_max = hist_max if (hist_max is not None) else quartiles["max"]

    # min value of histogram: either given as input or determined by dp min
    hist_min = hist_min if (hist_min is not None) else quartiles["min"]

    # if all values above defined max, create one histogram bin greater hist_max
    if hist_min > hist_max:
        bins = np.array([hist_max, np.inf])
        counts = np.array([len(series)])

    # if there are max. 10 integers, use value counts of single integers instead of bin ranges for histogram
    elif (bin_type is int) and (
        (bin_range == 1) or ((bin_range is None) and (hist_max - hist_min < 10))
    ):
        bins = np.array(range(int(quartiles["min"]), int(quartiles["max"]) + 1))
        counts = np.bincount(series, minlength=int(quartiles["max"]) + 1)[
            int(quartiles["min"]) :
        ]

        # sum all counts above hist max to single bin >max
        if bins[-1] > hist_max:
            bins = bins[bins <= hist_max]
            bins = np.append(bins, np.Inf)
            counts[len(bins) - 1] = counts[
                len(bins) - 1 :
            ].sum()  # sum up to single > max count
            counts = counts[: len(bins)]  # remove long tail

    # else use ranges for bins to create histogram
    else:
        # if hist range is small (<1), set bin_range to 0.1 to prevent too fine-granular bins
        if bin_range is None and (hist_max - hist_min) < 1:
            bin_range = 0.1

        # if bin_range is provided, snap min and max value accordingly(to create "clean" bins).
        if bin_range is not None:
            # "snap" min_value: e.g., if bin range is 5 and dp min value is 12, min_value snaps to 10
            hist_min = hist_min - (hist_min % bin_range)

        # set default of 10 bins if no bin_range is given.
        # compute bin_range from hist_max and snap max_value accordingly. This is necessary, bc hist_max and dp max might differ:
        # to create 10 bins up to hist_max but maintain bins up to dp max (to aggregate them in the following step)
        else:
            bin_range = (hist_max - hist_min) / 10

        # "snap" hist_max_input to bin_range (for pretty hist bins): E.g., if bin range is 5, and the dp max value is 23, max_value snaps to 25
        hist_max_input = hist_max if hist_max > quartiles["max"] else quartiles["max"]
        hist_max_input = (
            hist_max_input
            if round((hist_max_input - hist_min) % bin_range, 2) == 0
            else hist_max_input
            + (bin_range - ((hist_max_input - hist_min) % bin_range))
        )
        max_bins = int((hist_max_input - hist_min) / bin_range)
        max_bins = (
            max_bins if max_bins > 2 else 2
        )  # make sure there are at least 2 bins for histogram function to work

        hist = np.histogram(series, bins=max_bins, range=(hist_min, hist_max_input))
        counts = hist[0]
        bins = hist[1].astype(bin_type)

        # sum all counts above hist max to single bin >max
        if bins[-1] > hist_max:
            bins = bins[bins <= hist_max]
            counts[len(bins) - 1] = counts[
                len(bins) - 1 :
            ].sum()  # sum up to single > max count
            counts = counts[: len(bins)]  # remove long tail
            bins = np.append(bins, np.Inf)

        # as hist_min is currently only used for mobility_entropy with min_value = 0, we dont need to sum all counts below min accordingly
        # to be generically applicable, this would be needed!

    dp_counts = diff_privacy.counts_dp(counts, epsi, sensitivity)

    # set counts above dp_max(i.e, quartiles["max"]) to 0 (only so that bins are shown according to user input, even if they are empty)
    # if bin is "inf" it should be maintained as it is the aggregation of everything > hist_max and <= quartiles["max"]
    temp_bins_max = bins if len(bins) == len(dp_counts) else bins[:-1]
    dp_counts[
        np.logical_and(
            (temp_bins_max > quartiles["max"]), (np.not_equal(temp_bins_max, np.inf))
        )
    ] = 0
    # set counts below dp_min(i.e, quartiles["min"]) to 0
    # only needed for mobility_entropy (to show all possible bins even if below DP min)
    temp_bins_min = bins if len(bins) == len(dp_counts) else bins[1:]
    dp_counts[temp_bins_min < quartiles["min"]] = 0

    moe_laplace = diff_privacy.laplace_margin_of_error(0.95, epsi, sensitivity)

    # as percent instead of counts
    trip_counts = sum(dp_counts)
    if trip_counts > 0:
        dp_counts = dp_counts / trip_counts * 100
        moe_laplace = moe_laplace / trip_counts * 100

    return TupleSection(
        data=(dp_counts, bins),
        privacy_budget=eps,
        quartiles=quartiles,
        margin_of_error_laplace=moe_laplace,
        margin_of_error_expmech=moe_expmech,
    )


def get_epsi(evalu: bool, eps: Optional[float], elements: int) -> Optional[float]:
    if evalu or eps is None:
        return eps
    else:
        return eps / elements


def _cumsum(array: np.array) -> np.array:
    array[::-1].sort()
    return (array.cumsum() / array.sum()).round(2)


def cumsum(
    counts: np.array, eps: float, sensitivity: int, nsim: int = 10, nrow: int = 100
) -> DataFrame:
    df_cumsum = DataFrame()
    df_cumsum["n"] = np.arange(1, len(counts) + 1)
    df_cumsum["cum_perc"] = _cumsum(counts)

    # reuduce df size by only keeping max 100 values
    if len(df_cumsum) > nrow:
        nth = len(df_cumsum) // nrow
        last_row = df_cumsum.iloc[len(df_cumsum) - 1, :]
        df_cumsum = df_cumsum.iloc[::nth, :]
        # append last row
        if int(last_row.n) not in list(df_cumsum.n):
            pd.concat([df_cumsum, pd.DataFrame(last_row).T])

    df_cumsum.reset_index(drop=True, inplace=True)
    return df_cumsum



import os
import shutil
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Tuple

import pkg_resources  # type: ignore
from tqdm.auto import tqdm

if TYPE_CHECKING:
    from dp_mobility_report import DpMobilityReport, BenchmarkReport

from dp_mobility_report import constants as const
from dp_mobility_report.report.html import (
    config_templates,
    html_utils,
    od_analysis_templates,
    overview_templates,
    place_analysis_templates,
    user_analysis_templates,
)


def render_html(
    dpmreport: "DpMobilityReport",
    output_filename: str,
    top_n_flows: int = 100,
    disable_progress_bar: bool = False,
) -> Tuple[str, Path]:
    template_structure = html_utils.get_template("structure.html")
    temp_map_folder = Path(os.path.join(tempfile.gettempdir(), "maps"))

    # remove any old temp files in case there exist any
    shutil.rmtree(temp_map_folder, ignore_errors=True)
    os.mkdir(temp_map_folder)

    args: dict = {}

    args["title"] = "DP Mobility Report"
    args["output_filename"] = output_filename

    with tqdm(  # progress bar
        total=4, desc="Create HTML Output", disable=disable_progress_bar
    ) as pbar:

        args["config_segment"] = config_templates.render_config(dpmreport)

        if not set(const.OVERVIEW_ELEMENTS).issubset(dpmreport.analysis_exclusion):
            args["overview_segment"] = overview_templates.render_overview(dpmreport)
        pbar.update()

        if not set(const.PLACE_ELEMENTS).issubset(dpmreport.analysis_exclusion):
            args[
                "place_analysis_segment"
            ] = place_analysis_templates.render_place_analysis(
                dpmreport, temp_map_folder, output_filename
            )
        pbar.update()

        if not set(const.OD_ELEMENTS).issubset(dpmreport.analysis_exclusion):
            args["od_analysis_segment"] = od_analysis_templates.render_od_analysis(
                dpmreport, top_n_flows, temp_map_folder, output_filename
            )
        pbar.update()

        if not set(const.USER_ELEMENTS).issubset(dpmreport.analysis_exclusion):
            args[
                "user_analysis_segment"
            ] = user_analysis_templates.render_user_analysis(dpmreport)
        pbar.update()

    return (
        template_structure.render(args),
        temp_map_folder,
    )


def render_benchmark_html(
    benchmarkreport: "BenchmarkReport",
    output_filename: str,
    top_n_flows: int = 100,
    disable_progress_bar: bool = False,
) -> Tuple[str, Path]:
    template_structure = html_utils.get_template("structure.html")
    temp_map_folder = Path(os.path.join(tempfile.gettempdir(), "maps"))

    # remove any old temp files in case there exist any
    shutil.rmtree(temp_map_folder, ignore_errors=True)
    os.mkdir(temp_map_folder)

    args: dict = {}

    args["title"] = "DP Mobility Report: Benchmark"
    args["output_filename"] = output_filename

    with tqdm(  # progress bar
        total=4, desc="Create HTML Output", disable=disable_progress_bar
    ) as pbar:

        args["config_segment"] = config_templates.render_benchmark_config(
            benchmarkreport
        )

        if not set(const.OVERVIEW_ELEMENTS).issubset(
            benchmarkreport.analysis_exclusion
        ):
            args["overview_segment"] = overview_templates.render_benchmark_overview(
                benchmarkreport
            )
        pbar.update()

        if not set(const.PLACE_ELEMENTS).issubset(benchmarkreport.analysis_exclusion):
            args[
                "place_analysis_segment"
            ] = place_analysis_templates.render_benchmark_place_analysis(
                benchmarkreport, temp_map_folder, output_filename
            )
        pbar.update()

        if not set(const.OD_ELEMENTS).issubset(benchmarkreport.analysis_exclusion):
            args[
                "od_analysis_segment"
            ] = od_analysis_templates.render_benchmark_od_analysis(
                benchmarkreport, top_n_flows, temp_map_folder, output_filename
            )
        pbar.update()

        if not set(const.USER_ELEMENTS).issubset(benchmarkreport.analysis_exclusion):
            args[
                "user_analysis_segment"
            ] = user_analysis_templates.render_benchmark_user_analysis(benchmarkreport)
        pbar.update()

    return (
        template_structure.render(args),
        temp_map_folder,
    )


def create_html_assets(output_file: Path) -> None:
    path = Path(os.path.join(output_file, "assets"))
    if path.is_dir():
        shutil.rmtree(path)
    os.mkdir(path)

    asset_folder = pkg_resources.resource_filename(
        "dp_mobility_report", "report/html/html_templates/assets/"
    )

    for file_name in os.listdir(asset_folder):
        # only copy .css files (e.g., not __init__.py)
        if str.endswith(file_name, ".css"):
            # construct full file path
            source = os.path.join(asset_folder, file_name)
            destination = os.path.join(path, file_name)
            # copy only files
            if os.path.isfile(source):
                shutil.copy(source, destination)


def create_maps_folder(temp_map_folder: Path, output_dir: Path) -> None:
    path = Path(os.path.join(output_dir, "maps"))
    if path.is_dir():
        shutil.rmtree(path)
    os.makedirs(path)

    for file_name in os.listdir(temp_map_folder):
        # construct full file path
        source = os.path.join(temp_map_folder, file_name)
        destination = os.path.join(path, file_name)
        # copy only files
        if os.path.isfile(source):
            shutil.copy(source, destination)



from pyproj import CRS

# ANALYSIS (SEGMENTS AND ELEMENTS) NAMES
# segments
ALL = "all"  # deprecated
OVERVIEW = "overview"
PLACE_ANALYSIS = "place_analysis"
OD_ANALYSIS = "od_analysis"
USER_ANALYSIS = "user_analysis"
SEGMENTS = [OVERVIEW, PLACE_ANALYSIS, OD_ANALYSIS, USER_ANALYSIS]

# elements
DS_STATISTICS = "ds_statistics"
MISSING_VALUES = "missing_values"
TRIPS_OVER_TIME = "trips_over_time"
TRIPS_PER_WEEKDAY = "trips_per_weekday"
TRIPS_PER_HOUR = "trips_per_hour"
VISITS_PER_TILE = "visits_per_tile"
VISITS_PER_TIME_TILE = "visits_per_time_tile"
VISITS_PER_TILE_OUTLIERS = "visits_per_tile_outliers"
VISITS_PER_TILE_RANKING = "visits_per_tile_ranking"
VISITS_PER_TILE_QUARTILES = "visits_per_tile_quartiles"
OD_FLOWS = "od_flows"
OD_FLOWS_RANKING = "od_flows_ranking"
OD_FLOWS_QUARTILES = "od_flows_quartiles"
TRAVEL_TIME = "travel_time"
TRAVEL_TIME_QUARTILES = "travel_time_quartiles"
JUMP_LENGTH = "jump_length"
JUMP_LENGTH_QUARTILES = "jump_length_quartiles"
TRIPS_PER_USER = "trips_per_user"
TRIPS_PER_USER_QUARTILES = "trips_per_user_quartiles"
USER_TIME_DELTA = "user_time_delta"
USER_TIME_DELTA_QUARTILES = "user_time_delta_quartiles"
RADIUS_OF_GYRATION = "radius_of_gyration"
RADIUS_OF_GYRATION_QUARTILES = "radius_of_gyration_quartiles"
USER_TILE_COUNT = "user_tile_count"
USER_TILE_COUNT_QUARTILES = "user_tile_count_quartiles"
MOBILITY_ENTROPY = "mobility_entropy"
MOBILITY_ENTROPY_QUARTILES = "mobility_entropy_quartiles"
OVERVIEW_ELEMENTS = [
    DS_STATISTICS,
    MISSING_VALUES,
    TRIPS_OVER_TIME,
    TRIPS_PER_WEEKDAY,
    TRIPS_PER_HOUR,
]
PLACE_ELEMENTS = [VISITS_PER_TILE, VISITS_PER_TIME_TILE]
OD_ELEMENTS = [OD_FLOWS, TRAVEL_TIME, JUMP_LENGTH]
USER_ELEMENTS = [
    TRIPS_PER_USER,
    USER_TIME_DELTA,
    RADIUS_OF_GYRATION,
    USER_TILE_COUNT,
    MOBILITY_ENTROPY,
]

ELEMENTS = OVERVIEW_ELEMENTS + PLACE_ELEMENTS + OD_ELEMENTS + USER_ELEMENTS
TESSELLATION_ELEMENTS = PLACE_ELEMENTS + [OD_FLOWS, USER_TILE_COUNT, MOBILITY_ENTROPY]
TIMESTAMP_ANALYSES = [
    TRIPS_OVER_TIME,
    TRIPS_PER_WEEKDAY,
    TRIPS_PER_HOUR,
    VISITS_PER_TIME_TILE,
    TRAVEL_TIME,
    USER_TIME_DELTA,
]

SEGMENTS_AND_ELEMENTS = SEGMENTS + ELEMENTS

# DATASET COLUMNS
ID = "id"
UID = "uid"
TID = "tid"
LAT = "lat"
LNG = "lng"
DATETIME = "datetime"

POINT_TYPE = "point_type"
START = "start"
END = "end"
TILE_ID = "tile_id"
TILE_NAME = "tile_name"
GEOMETRY = "geometry"

TILE_ID_END = TILE_ID + "_" + END
LNG_END = LNG + "_" + END
LAT_END = LAT + "_" + END
DATETIME_END = DATETIME + "_" + END

# TIME
DATE = "date"
DAY_NAME = "day_name"
HOUR = "hour"
IS_WEEKEND = "is_weekend"
WEEKDAY = "weekday"
WEEKEND = "weekend"
TIME_CATEGORY = "time category"


# OTHER
DEFAULT_CRS = CRS.from_epsg(4326)
QUARTILES = "quartiles"
DATETIME_PRECISION = "datetime_precision"
PREC_DATE = "date"
PREC_WEEK = "week"
PREC_MONTH = "month"

ORIGIN = "origin"
DESTINATION = "destination"
FLOW = "flow"

# ds_statistics elements
N_RECORDS = "n_records"
N_TRIPS = "n_trips"
N_COMPLETE_TRIPS = "n_complete_trips"
N_INCOMPLETE_TRIPS = "n_incomplete_trips"
N_USERS = "n_users"
N_LOCATIONS = "n_locations"
DS_STATISTICS_ELEMENTS = [
    N_RECORDS,
    N_TRIPS,
    N_COMPLETE_TRIPS,
    N_INCOMPLETE_TRIPS,
    N_USERS,
    N_LOCATIONS,
]

# missing_values elements
MISSING_VALUES_ELEMENTS = [UID, TID, DATETIME, LAT, LNG]

# SIMILARITY MEASURES
SMAPE = "symmetric_mean_absolute_percentage_error"
JSD = "jensen_shannon_divergence"
KLD = "kullback_leibler_divergence"
EMD = "earth_movers_distance"
KT = "kendalls_tau"
TOP_N_COV = "top_n_coverage"
SIMILARITY_MEASURES = [SMAPE, JSD, KLD, EMD, KT, TOP_N_COV]

DISABLE_EMD_THRESHOLD = 1000

DARK_BLUE = "#283377"
LIGHT_BLUE = "#5D6FFF"
ORANGE = "#D9642C"
LIGHT_ORANGE = "#FFAD6F"
GREY = "#8A8A8A"
LIGHT_GREY = "#f2f2f2"
DARK_RED = "#8B0000"
LIGHT_RED = "#FFCCCB"

STANDARD_CMAP = "YlOrRd"
BASE_CMAP = "Blues"
ALT_CMAP = "Oranges"
DIVERGING_CMAP = "PiYG"



from typing import TYPE_CHECKING

import pandas as pd

import dp_mobility_report.constants as const
from dp_mobility_report.report.html.html_utils import fmt, fmt_config, get_template

if TYPE_CHECKING:
    from dp_mobility_report import BenchmarkReport, DpMobilityReport


def render_config(dpmreport: "DpMobilityReport") -> str:

    args: dict = {}

    args["config_table"] = render_config_table(dpmreport)
    args["privacy_info"] = render_privacy_info(dpmreport.privacy_budget is not None)
    if dpmreport.tessellation is None:
        args[
            "tessellation_info"
        ] = "No tessellation has been provided. All analyses based on the tessellation have been excluded."

    if not pd.core.dtypes.common.is_datetime64_dtype(dpmreport.df[const.DATETIME]):
        args[
            "timestamp_info"
        ] = "Dataframe does not contain timestamps. All analyses based on timestamps have been excluded."

    if max(dpmreport.df[const.TID].value_counts()) == 1:
        args[
            "od_info"
        ] = "All trips in the dataset only contain a single record, therefore, all origin-destination analyses (OD Flows, travel time, jump length) have been excluded."

    template_structure = get_template("config_segment.html")
    return template_structure.render(args)


def render_benchmark_config(benchmarkreport: "BenchmarkReport") -> str:

    args: dict = {}

    args["config_table"] = render_benchmark_config_table(benchmarkreport)

    if benchmarkreport.report_base.tessellation is None:
        args[
            "tessellation_info"
        ] = "No tessellation has been provided. All analyses based on the tessellation have been excluded."

    if (
        not pd.core.dtypes.common.is_datetime64_dtype(
            benchmarkreport.report_base.df[const.DATETIME]
        )
    ) or (
        not pd.core.dtypes.common.is_datetime64_dtype(
            benchmarkreport.report_alternative.df[const.DATETIME]
        )
    ):
        args[
            "timestamp_info"
        ] = "At least one of the datasets does not contain timestamps. All analyses based on timestamps have been excluded."

    if (max(benchmarkreport.report_base.df[const.TID].value_counts()) == 1) or (
        max(benchmarkreport.report_alternative.df[const.TID].value_counts()) == 1
    ):
        args[
            "od_info"
        ] = "All trips in at least one of the datasets only contain a single record, therefore, all origin-destination analyses (OD Flows, travel time, jump length) have been excluded."

    template_structure = get_template("config_segment.html")
    return template_structure.render(args)


def render_config_table(dpmreport: "DpMobilityReport") -> str:

    config_list = [
        {"name": "Max. trips per user", "value": fmt(dpmreport.max_trips_per_user)},
        {"name": "Privacy budget", "value": fmt(dpmreport.privacy_budget)},
        {"name": "User privacy", "value": fmt(dpmreport.user_privacy)},
        {
            "name": "Excluded analyses",
            "value": fmt_config(dpmreport.analysis_exclusion),
        },
        {"name": "Budget split", "value": fmt_config(dpmreport.budget_split)},
        {"name": "Evaluation dev. mode", "value": fmt(dpmreport.evalu)},
    ]

    # create html from template
    template_table = get_template("table.html")
    dataset_stats_html = template_table.render(rows=config_list)
    return dataset_stats_html


def render_benchmark_config_table(benchmarkreport: "BenchmarkReport") -> str:

    config_list = [
        {
            "name": "Max. trips per user",
            "value": (
                fmt(benchmarkreport.report_base.max_trips_per_user),
                fmt(benchmarkreport.report_alternative.max_trips_per_user),
            ),
        },
        {
            "name": "Privacy budget",
            "value": (
                fmt(benchmarkreport.report_base.privacy_budget),
                fmt(benchmarkreport.report_alternative.privacy_budget),
            ),
        },
        {
            "name": "User privacy",
            "value": (
                fmt(benchmarkreport.report_base.user_privacy),
                fmt(benchmarkreport.report_alternative.privacy_budget),
            ),
        },
        {
            "name": "Budget split",
            "value": (
                fmt_config(benchmarkreport.report_base.budget_split),
                fmt_config(benchmarkreport.report_alternative.budget_split),
            ),
        },
        {
            "name": "Evaluation dev. mode",
            "value": (
                fmt(benchmarkreport.report_base.evalu),
                fmt(benchmarkreport.report_alternative.evalu),
            ),
        },
        {
            "name": "Excluded analyses",
            "value": (
                fmt_config(benchmarkreport.analysis_exclusion),
                fmt_config(benchmarkreport.analysis_exclusion),
            ),
        },  # TODO verbundene zelle
    ]

    # create html from template
    template_table = get_template("table_benchmark.html")
    dataset_stats_html = template_table.render(name="Configuration", rows=config_list)
    return dataset_stats_html


def render_privacy_info(with_privacy: bool) -> str:
    if with_privacy:
        return "Noise has been added to provide differential privacy. The 95%-confidence interval gives an intuition about the reliability of the noisy results."
    else:
        return "No noise has been added."



import calendar
from datetime import timedelta
from typing import TYPE_CHECKING, Optional

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from dp_mobility_report import DpMobilityReport

from dp_mobility_report import constants as const
from dp_mobility_report.model import m_utils
from dp_mobility_report.model.section import DfSection, DictSection, SeriesSection
from dp_mobility_report.privacy import diff_privacy


def get_dataset_statistics(
    dpmreport: "DpMobilityReport", eps: Optional[float]
) -> DictSection:
    epsi = m_utils.get_epsi(dpmreport.evalu, eps, 4)

    # counts for complete and incomplete trips
    points_per_trip = (
        dpmreport.df.reset_index().groupby(const.TID).count()["index"].value_counts()
    )
    n_incomplete_trips = 0 if 1 not in points_per_trip else points_per_trip[1]
    n_incomplete_trips = diff_privacy.count_dp(
        n_incomplete_trips,
        epsi,
        dpmreport.count_sensitivity_base,
    )

    moe_incomplete_trips = diff_privacy.laplace_margin_of_error(
        0.95, epsi, dpmreport.count_sensitivity_base
    )

    n_complete_trips = 0 if 2 not in points_per_trip else points_per_trip[2]
    n_complete_trips = diff_privacy.count_dp(
        n_complete_trips,
        epsi,
        2 * dpmreport.count_sensitivity_base,
    )
    moe_complete_trips = diff_privacy.laplace_margin_of_error(
        0.95, epsi, 2 * dpmreport.count_sensitivity_base
    )

    n_trips = n_incomplete_trips + n_complete_trips
    if n_trips == 0:
        n_trips = None  # trips cannot be None
        moe_trips = (moe_incomplete_trips + moe_complete_trips) / 2
    else:
        moe_trips = (
            n_incomplete_trips * moe_incomplete_trips
            + n_complete_trips * moe_complete_trips
        ) / n_trips

    n_records = None if n_trips is None else (n_incomplete_trips + n_complete_trips * 2)
    if n_records is None:
        moe_records = (moe_incomplete_trips + moe_complete_trips * 2) / 2
    else:
        moe_records = (
            n_incomplete_trips * moe_incomplete_trips
            + n_complete_trips * 2 * moe_complete_trips
        ) / n_records

    n_users = diff_privacy.count_dp(
        dpmreport.df[const.UID].nunique(), epsi, 1, nonzero=True
    )
    moe_users = diff_privacy.laplace_margin_of_error(0.95, epsi, 1)

    n_locations = diff_privacy.count_dp(
        dpmreport.df.groupby([const.LAT, const.LNG]).ngroups,
        epsi,
        2 * dpmreport.count_sensitivity_base,
        nonzero=True,
    )
    moe_locations = diff_privacy.laplace_margin_of_error(
        0.95, epsi, 2 * dpmreport.count_sensitivity_base
    )

    stats = {
        const.N_RECORDS: n_records,
        const.N_TRIPS: n_trips,
        const.N_COMPLETE_TRIPS: n_complete_trips,
        const.N_INCOMPLETE_TRIPS: n_incomplete_trips,
        const.N_USERS: n_users,
        const.N_LOCATIONS: n_locations,
    }

    moe = {
        const.N_RECORDS: moe_records,
        const.N_TRIPS: moe_trips,
        const.N_COMPLETE_TRIPS: moe_complete_trips,
        const.N_INCOMPLETE_TRIPS: moe_incomplete_trips,
        const.N_USERS: moe_users,
        const.N_LOCATIONS: moe_locations,
    }

    return DictSection(data=stats, privacy_budget=eps, margin_of_errors_laplace=moe)


def get_missing_values(
    dpmreport: "DpMobilityReport", eps: Optional[float]
) -> DictSection:
    columns = [const.UID, const.TID, const.DATETIME, const.LAT, const.LNG]
    epsi = m_utils.get_epsi(dpmreport.evalu, eps, len(columns))

    missings = dict((len(dpmreport.df) - dpmreport.df.count())[columns])

    moe = diff_privacy.laplace_margin_of_error(
        0.95, epsi, 2 * dpmreport.count_sensitivity_base
    )
    conf_interval = {}
    for col in columns:
        missings[col] = diff_privacy.count_dp(
            missings[col], epsi, 2 * dpmreport.count_sensitivity_base
        )
        conf_interval["ci95_" + col] = diff_privacy.conf_interval(missings[col], moe)

    return DictSection(data=missings, privacy_budget=eps, margin_of_error_laplace=moe)


def get_trips_over_time(
    dpmreport: "DpMobilityReport", eps: Optional[float]
) -> DfSection:
    epsi = m_utils.get_epsi(dpmreport.evalu, eps, 3)
    epsi_limits = epsi * 2 if epsi is not None else None

    df_trip = dpmreport.df[
        (dpmreport.df[const.POINT_TYPE] == const.END)
    ]  # only count each trip once
    dp_bounds = diff_privacy.bounds_dp(
        df_trip[const.DATETIME], epsi_limits, dpmreport.count_sensitivity_base
    )

    # cut based on dp min and max values
    (trips_over_time) = pd.DataFrame(
        m_utils.cut_outliers(  # don't disclose outliers to the as the boundaries are not defined through user input
            df_trip[const.DATETIME],
            min_value=dp_bounds[0],
            max_value=dp_bounds[1],
        )
    )

    # only use date and remove time
    dp_bounds = pd.Series(dp_bounds).dt.date

    # different aggregations based on range of dates
    range_of_days = dp_bounds[1] - dp_bounds[0]
    if range_of_days > timedelta(days=712):  # more than two years (102 weeks)
        resample = "M"
        datetime_precision = const.PREC_MONTH
    if range_of_days > timedelta(days=90):  # more than three months
        resample = "W-Mon"
        datetime_precision = const.PREC_WEEK
    else:
        resample = "1D"
        datetime_precision = const.PREC_DATE

    trips_over_time.loc[:, "trip_count"] = 1
    trips_over_time = (
        trips_over_time.set_index(const.DATETIME)
        .resample(resample, label="left")
        .count()
        .reset_index()
    )
    trips_over_time[const.DATETIME] = trips_over_time[const.DATETIME].dt.date
    trips_over_time["trip_count"] = diff_privacy.counts_dp(
        trips_over_time["trip_count"].values,
        epsi,
        dpmreport.count_sensitivity_base,
    )

    moe_laplace = diff_privacy.laplace_margin_of_error(
        0.95, epsi, dpmreport.count_sensitivity_base
    )

    # as percent instead of absolute values
    trip_sum = np.sum(trips_over_time["trip_count"])
    if trip_sum != 0:
        trips_over_time["trips"] = trips_over_time["trip_count"] / trip_sum * 100
        moe_laplace = moe_laplace / trip_sum * 100
    else:
        trips_over_time["trips"] = 0

    quartiles = pd.Series({"min": dp_bounds[0], "max": dp_bounds[1]})

    return DfSection(
        data=trips_over_time,
        privacy_budget=eps,
        datetime_precision=datetime_precision,
        quartiles=quartiles,
        margin_of_error_laplace=moe_laplace,
    )


def get_trips_per_weekday(
    dpmreport: "DpMobilityReport", eps: Optional[float]
) -> SeriesSection:
    dpmreport.df.loc[:, const.DATE] = dpmreport.df[const.DATETIME].dt.date
    dpmreport.df.loc[:, const.DAY_NAME] = dpmreport.df[const.DATETIME].dt.day_name()
    dpmreport.df.loc[:, const.WEEKDAY] = dpmreport.df[const.DATETIME].dt.weekday

    trips_per_weekday = (
        dpmreport.df[
            dpmreport.df[const.POINT_TYPE] == const.END
        ]  # count trips not records
        .groupby([const.DAY_NAME], sort=False)
        .count()[const.TID]
    )
    missing_days = set(calendar.day_name) - set(trips_per_weekday.index.tolist())
    trips_per_weekday = pd.concat(
        [trips_per_weekday, pd.Series(0, index=list(missing_days))]
    )

    trips_per_weekday.index = pd.Categorical(
        trips_per_weekday.index, list(calendar.day_name)
    )
    trips_per_weekday.sort_index(inplace=True)

    trips_per_weekday = pd.Series(
        index=trips_per_weekday.index,
        data=diff_privacy.counts_dp(
            trips_per_weekday.values,
            eps,
            dpmreport.count_sensitivity_base,
        ),
    )
    moe = diff_privacy.laplace_margin_of_error(
        0.95, eps, dpmreport.count_sensitivity_base
    )

    # as percent instead of absolute values
    trip_sum = np.sum(trips_per_weekday)
    if trip_sum != 0:
        trips_per_weekday = trips_per_weekday / trip_sum * 100
        moe = moe / trip_sum * 100

    return SeriesSection(
        data=trips_per_weekday, privacy_budget=eps, margin_of_error_laplace=moe
    )


def get_trips_per_hour(
    dpmreport: "DpMobilityReport", eps: Optional[float]
) -> DfSection:
    hour_weekday = dpmreport.df.groupby(
        [const.HOUR, const.IS_WEEKEND, const.POINT_TYPE]
    ).count()[const.TID]
    hour_weekday.name = "count"

    # create all potential combinations
    full_combination = pd.DataFrame(
        list(
            map(
                np.ravel,
                np.meshgrid(
                    range(0, 24),
                    [const.WEEKDAY, const.WEEKEND],
                    [const.START, const.END],
                ),
            )
        )
    ).T
    full_combination.columns = [const.HOUR, const.IS_WEEKEND, const.POINT_TYPE]
    full_combination["count"] = 0
    full_combination.set_index(
        [const.HOUR, const.IS_WEEKEND, const.POINT_TYPE], inplace=True
    )

    hour_weekday = full_combination.add(pd.DataFrame(hour_weekday), fill_value=0)

    hour_weekday = hour_weekday.reset_index()
    hour_weekday["count"] = diff_privacy.counts_dp(
        hour_weekday["count"], eps, dpmreport.count_sensitivity_base
    )

    hour_weekday[const.TIME_CATEGORY] = (
        hour_weekday[const.IS_WEEKEND] + " " + hour_weekday[const.POINT_TYPE]
    )
    moe = diff_privacy.laplace_margin_of_error(
        0.95, eps, dpmreport.count_sensitivity_base
    )

    # as percent instead of absolute values
    trip_sum = np.sum(
        hour_weekday[hour_weekday.point_type == const.END]["count"]
    )  # only use ends to get sum of trips
    if trip_sum != 0:
        hour_weekday["perc"] = hour_weekday["count"] / trip_sum * 100
        moe = moe / trip_sum * 100

    return DfSection(
        data=hour_weekday[[const.HOUR, const.TIME_CATEGORY, "perc"]],
        privacy_budget=eps,
        margin_of_error_laplace=moe,
    )



import os
import warnings
from pathlib import Path
from shutil import rmtree
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from pandarallel import pandarallel
from pandas import DataFrame
from tqdm.auto import tqdm

from dp_mobility_report import constants as const
from dp_mobility_report.model import preprocessing
from dp_mobility_report.report import report
from dp_mobility_report.report.html.templates import (
    create_html_assets,
    create_maps_folder,
    render_html,
)


class DpMobilityReport:
    """Generate a (differentially private) mobility report from a mobility dataset. The report will be generated as an HTML file, using the ``.to_file()`` method.
    Args:
        df: ``DataFrame`` containing the mobility data. Expected columns: User ID ``uid``, trip ID ``tid``, timestamp ``datetime`` (or ``int`` to indicate sequence position, if dataset only consists of sequences without timestamps), latitude ``lat`` and longitude ``lng`` in CRS EPSG:4326.
        tessellation: Geopandas ``GeoDataFrame`` containing the tessellation for spatial aggregations. Expected columns: ``tile_id``. If tessellation is not provided in the expected default CRS EPSG:4326, it will automatically be transformed. If no tessellation is provided, all analyses based on the tessellation will automatically be removed.
        privacy_budget: privacy_budget for the differentially private report. Defaults to ``None``, i.e., no privacy guarantee is provided.
        user_privacy: Whether item-level or user-level privacy is applied. Defaults to ``True`` (user-level privacy).
        max_trips_per_user: Maximum number of trips a user is allowed to contribute to the data. Dataset will be sampled accordingly. Defaults to ``None``, i.e., all trips are used. This implies that the actual maximum number of trips per user will be used according to the data, though this violates user-level Differential Privacy.
        analysis_selection: Select only needed analyses. A selection reduces computation time and leaves more privacy budget
        for higher accuracy of other analyses. ``analysis_selection`` takes a list of all analyses to be included. Alternatively, a list of analyses to be excluded can be set with ``analysis_exclusion``.
        Either entire segments can be included: ``const.OVERVIEW``, ``const.PLACE_ANALYSIS``, ``const.OD_ANALYSIS``, ``const.USER_ANALYSIS``
        or any single analysis can be included: ``const.DS_STATISTICS``, ``const.MISSING_VALUES``, ``const.TRIPS_OVER_TIME``, ``const.TRIPS_PER_WEEKDAY``, ``const.TRIPS_PER_HOUR``, ``const.VISITS_PER_TILE``, ``const.VISITS_PER_TIME_TILE``, ``const.OD_FLOWS``, ``const.TRAVEL_TIME``, ``const.JUMP_LENGTH``, ``const.TRIPS_PER_USER``, ``const.USER_TIME_DELTA``, ``const.RADIUS_OF_GYRATION``, ``const.USER_TILE_COUNT``, ``const.MOBILITY_ENTROPY``
        Default is None, i.e., all analyses are included.
        analysis_exclusion: Ignored, if ``analysis_selection`` is set! ``analysis_exclusion`` takes a list of all analyses to be excluded.
        Either entire segments can be excluded: ``const.OVERVIEW``, ``const.PLACE_ANALYSIS``, ``const.OD_ANALYSIS``, ``const.USER_ANALYSIS``
        or any single analysis can be excluded: ``const.DS_STATISTICS``, ``const.MISSING_VALUES``, ``const.TRIPS_OVER_TIME``, ``const.TRIPS_PER_WEEKDAY``, ``const.TRIPS_PER_HOUR``, ``const.VISITS_PER_TILE``, ``const.VISITS_PER_TIME_TILE``, ``const.OD_FLOWS``, ``const.TRAVEL_TIME``, ``const.JUMP_LENGTH``, ``const.TRIPS_PER_USER``, ``const.USER_TIME_DELTA``, ``const.RADIUS_OF_GYRATION``, ``const.USER_TILE_COUNT``, ``const.MOBILITY_ENTROPY``
        budget_split: ``dict`` to customize how much privacy budget is assigned to which analysis. Each key needs to be named according to an analysis and the value needs to be an integer indicating the weight for the privacy budget.
        If no weight is assigned, a default weight of 1 is set.
        For example, if ``budget_split = {const.VISITS_PER_TILE: 10}``, then the privacy budget for ``visits_per_tile`` is 10 times higher than for every other analysis, which all get a default weight of 1.
        Possible ``dict`` keys (all analyses): ``const.DS_STATISTICS``, ``const.MISSING_VALUES``, ``const.TRIPS_OVER_TIME``, ``const.TRIPS_PER_WEEKDAY``, ``const.TRIPS_PER_HOUR``, ``const.VISITS_PER_TILE``, ``const.VISITS_PER_TIME_TILE``, ``const.OD_FLOWS``, ``const.TRAVEL_TIME``, ``const.JUMP_LENGTH``, ``const.TRIPS_PER_USER``, ``const.USER_TIME_DELTA``, ``const.RADIUS_OF_GYRATION``, ``const.USER_TILE_COUNT``, ``const.MOBILITY_ENTROPY``
        timewindows: List of hours as ``int`` that define the timewindows for the spatial analysis for single time windows. Defaults to ``[2, 6, 10, 14, 18, 22]``.
        max_travel_time: Upper bound for travel time histogram. If ``None`` is given, no upper bound is set. Defaults to ``None``.
        bin_range_travel_time: The range a single histogram bin spans for travel time (e.g., 5 for 5 min bins). If ``None`` is given, the histogram bins will be determined automatically. Defaults to ``None``.
        max_jump_length: Upper bound for jump length histogram. If ``None`` is given, no upper bound is set. Defaults to ``None``.
        bin_range_jump_length: The range a single histogram bin spans for jump length (e.g., 1 for 1 km bins). If ``None`` is given, the histogram bins will be determined automatically. Defaults to ``None``.
        max_radius_of_gyration: Upper bound for radius of gyration histogram. If ``None`` is given, no upper bound is set. Defaults to ``None``.
        bin_range_radius_of_gyration: The range a single histogram bin spans for the radius of gyration (e.g., 1 for 1 km bins). If ``None`` is given, the histogram bins will be determined automatically. Defaults to ``None``.
        max_user_tile_count: Upper bound for distinct tiles per user histogram. If ``None`` is given, no upper bound is set. Defaults to ``None``.
        bin_range_user_tile_count: The range a single histogram bin spans for the distinct tiles per user histogram. If ``None`` is given, the histogram bins will be determined automatically. Defaults to ``None``.
        max_user_time_delta:  Upper bound for user time delta histogram. If ``None`` is given, no upper bound is set. Defaults to ``None``.
        bin_range_user_time_delta: The range a single histogram bin spans for user time delta (e.g., 1 for 1 hour bins). If ``None`` is given, the histogram bins will be determined automatically. Defaults to ``None``.
        disable_progress_bar: Whether progress bars should be shown. Defaults to ``False``.
        seed_sampling: Provide seed for down-sampling of dataset (according to ``max_trips_per_user``) so that the sampling is reproducible. Defaults to ``None``, i.e., no seed.
        evalu (bool, optional): Parameter only needed for development and evaluation purposes. Defaults to ``False``."""

    _report: dict = {}
    _html: str = ""
    _df: DataFrame
    _tessellation: Optional[GeoDataFrame]
    _privacy_budget: Optional[Union[int, float]]
    _max_trips_per_user: int
    _analysis_exclusion: list
    _budget_split: dict

    def __init__(
        self,
        df: DataFrame,
        tessellation: Optional[GeoDataFrame] = None,
        privacy_budget: Optional[Union[int, float]] = None,
        user_privacy: bool = True,
        max_trips_per_user: Optional[int] = None,
        analysis_selection: Optional[List[str]] = None,
        analysis_exclusion: Optional[List[str]] = None,
        budget_split: dict = {},
        timewindows: Union[List[int], np.ndarray] = [2, 6, 10, 14, 18, 22],
        max_travel_time: Optional[int] = None,
        bin_range_travel_time: Optional[int] = None,
        max_jump_length: Optional[Union[int, float]] = None,
        bin_range_jump_length: Optional[Union[int, float]] = None,
        max_radius_of_gyration: Optional[Union[int, float]] = None,
        bin_range_radius_of_gyration: Optional[Union[int, float]] = None,
        max_user_tile_count: Optional[int] = None,
        bin_range_user_tile_count: Optional[int] = None,
        max_user_time_delta: Optional[Union[int, float]] = None,
        bin_range_user_time_delta: Optional[Union[int, float]] = None,
        disable_progress_bar: bool = False,
        seed_sampling: int = None,
        evalu: bool = False,
    ) -> None:
        preprocessing.validate_input(
            df,
            tessellation,
            privacy_budget,
            max_trips_per_user,
            analysis_selection,
            analysis_exclusion,
            budget_split,
            disable_progress_bar,
            evalu,
            user_privacy,
            timewindows,
            max_travel_time,
            bin_range_travel_time,
            max_jump_length,
            bin_range_jump_length,
            max_radius_of_gyration,
            bin_range_radius_of_gyration,
            max_user_tile_count,
            bin_range_user_tile_count,
            max_user_time_delta,
            bin_range_user_time_delta,
            seed_sampling,
        )

        (
            analysis_selection,
            analysis_exclusion,
        ) = preprocessing.validate_inclusion_exclusion(
            analysis_selection,
            analysis_exclusion,
        )

        self.user_privacy = user_privacy
        with tqdm(  # progress bar
            total=2, desc="Preprocess data", disable=disable_progress_bar
        ) as pbar:
            self._tessellation = (
                None
                if tessellation is None
                else preprocessing.preprocess_tessellation(tessellation)
            )
            pbar.update()

            self._max_trips_per_user = (
                max_trips_per_user
                if max_trips_per_user is not None
                else df.groupby(const.UID).nunique()[const.TID].max()
            )

            if user_privacy:
                self.count_sensitivity_base = self._max_trips_per_user
            else:
                self.count_sensitivity_base = 1
            self._df = preprocessing.preprocess_data(
                df.copy(),  # copy, to not overwrite users instance of df
                self.tessellation,
                self.max_trips_per_user,
                self.user_privacy,
                seed_sampling,
            )
            pbar.update()

        self._privacy_budget = None if privacy_budget is None else float(privacy_budget)
        self.max_travel_time = max_travel_time
        timewindows.sort()
        self.timewindows = (
            np.array(timewindows) if isinstance(timewindows, list) else timewindows
        )
        self.max_jump_length = max_jump_length
        self.bin_range_jump_length = bin_range_jump_length
        self.bin_range_travel_time = bin_range_travel_time
        self.max_radius_of_gyration = max_radius_of_gyration
        self.bin_range_radius_of_gyration = bin_range_radius_of_gyration
        self.max_user_tile_count = max_user_tile_count
        self.bin_range_user_tile_count = bin_range_user_tile_count
        self.max_user_time_delta = max_user_time_delta
        self.bin_range_user_time_delta = bin_range_user_time_delta
        self._analysis_exclusion = preprocessing.clean_analysis_exclusion(
            analysis_selection,
            analysis_exclusion,
            has_tessellation=(tessellation is not None),
            has_timestamps=pd.core.dtypes.common.is_datetime64_dtype(
                self.df[const.DATETIME]
            ),
            has_od_flows=max(self.df[const.TID].value_counts())
            > 1,  # are there trips with more than a single record?
            has_consecutive_user_trips=max(
                self.df.groupby(const.UID).nunique()[const.TID]
            )
            > 1,
        )
        self._budget_split = preprocessing.clean_budget_split(
            budget_split, self._analysis_exclusion
        )
        self.evalu = evalu
        self.disable_progress_bar = disable_progress_bar

        # initialize parallel processing
        pandarallel.initialize(verbose=0)

    @property
    def report(self) -> dict:
        """A dictionary with all report elements (i.e., analyses)."""
        if not self._report:
            self._report = report.report_elements(self)
        return self._report

    @property
    def df(self) -> DataFrame:
        """DataFrame containing the processed input mobility data of the report."""
        return self._df

    @property
    def max_trips_per_user(self) -> int:
        """Maximum number of trips per user as specified in the parameters. If ``None`` was given, this equals the actual maximum according to the data."""
        return self._max_trips_per_user

    @property
    def budget_split(self) -> dict:
        """Budget split as specified in the parameters."""
        return self._budget_split

    @property
    def analysis_exclusion(self) -> list:
        """List of analyses that have been excluded from the report and similarity measures. If analysis selection was provided as a parameter, they are inverted to this ``analysis_exclusion`` parameter."""
        return self._analysis_exclusion

    @property
    def privacy_budget(self) -> Union[int, float]:
        """Privacy budget as specified in the parameters."""
        return self._privacy_budget

    @property
    def tessellation(self) -> GeoDataFrame:
        """Processed tessellation."""
        return self._tessellation

    def to_file(
        self,
        output_file: Union[str, Path],
        disable_progress_bar: Optional[bool] = None,
        top_n_flows: int = 100,
    ) -> None:
        """Write the report to a file.
        By default a name is generated.

        Args:
            output_file: The name or the path of the file to store the ``html`` output.
            disable_progress_bar: if ``False``, no progress bar is shown.
            top_n_flows: Determines how many of the top ``n`` origin-destination flows are displayed. Defaults to 100.
        """
        if disable_progress_bar is None:
            disable_progress_bar = self.disable_progress_bar

        if not isinstance(output_file, Path):
            output_file = Path(str(output_file))

        else:
            if output_file.suffix != ".html":
                suffix = output_file.suffix
                output_file = output_file.with_suffix(".html")
                warnings.warn(
                    f"Extension {suffix} not supported. For now we assume .html was intended. "
                    f"To remove this warning, please use .html or .json."
                )

        output_dir = Path(os.path.splitext(output_file)[0])
        filename = Path(os.path.basename(output_file)).stem

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        create_html_assets(output_dir)

        # create report if not created yet (to display progress bar in correct order)
        self.report

        # render html
        data, temp_map_folder = render_html(
            self, filename, top_n_flows, disable_progress_bar
        )

        create_maps_folder(temp_map_folder, output_dir)

        # clean up temp folder
        rmtree(temp_map_folder, ignore_errors=True)

        output_file.write_text(data, encoding="utf-8")



import os
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from geopandas.geodataframe import GeoDataFrame

if TYPE_CHECKING:
    from dp_mobility_report import DpMobilityReport, BenchmarkReport

import folium

from dp_mobility_report import constants as const
from dp_mobility_report.benchmark.similarity_measures import symmetric_perc_error
from dp_mobility_report.model import od_analysis
from dp_mobility_report.model.section import DfSection, TupleSection
from dp_mobility_report.report.html.html_utils import (
    all_available_measures,
    fmt_moe,
    get_template,
    render_benchmark_summary,
    render_eps,
    render_moe_info,
    render_summary,
    render_user_input_info,
)
from dp_mobility_report.visualization import plot, v_utils


def render_od_analysis(
    dpmreport: "DpMobilityReport",
    top_n_flows: int,
    temp_map_folder: Path,
    output_filename: str,
) -> str:
    THRESHOLD = 0.2  # 20 %
    args: dict = {}
    report = dpmreport.report

    args[
        "user_config_info"
    ] = f"User configuration: display max. top {top_n_flows} OD connections on map"
    args["output_filename"] = output_filename

    if const.OD_FLOWS not in dpmreport.analysis_exclusion:
        args["od_eps"] = render_eps(report[const.OD_FLOWS].privacy_budget)
        args["od_moe"] = fmt_moe(report[const.OD_FLOWS].margin_of_error_laplace)

        args["od_legend"] = render_origin_destination_flows(
            report[const.OD_FLOWS],
            dpmreport.tessellation,
            top_n_flows,
            THRESHOLD,
            temp_map_folder,
        )
        args["intra_tile_flows_info"] = render_intra_tile_flows(
            report[const.OD_FLOWS], len(dpmreport.tessellation)
        )
        quartiles = report[const.OD_FLOWS].quartiles.round()
        args["flows_summary_table"] = render_summary(quartiles.astype(int))
        args["flows_cumsum_linechart"] = render_flows_cumsum(
            report[const.OD_FLOWS], diagonal=True
        )
        args["most_freq_flows_ranking"] = render_most_freq_flows_ranking(
            report[const.OD_FLOWS], dpmreport.tessellation
        )

    if const.TRAVEL_TIME not in dpmreport.analysis_exclusion:
        args["travel_time_eps"] = render_eps(report[const.TRAVEL_TIME].privacy_budget)
        args["travel_time_moe"] = fmt_moe(
            report[const.TRAVEL_TIME].margin_of_error_laplace
        )

        args["travel_time_hist_info"] = render_user_input_info(
            dpmreport.max_travel_time, dpmreport.bin_range_travel_time
        )
        args["travel_time_hist"] = render_travel_time_hist(report[const.TRAVEL_TIME])
        args["travel_time_moe_info"] = render_moe_info(
            report[const.TRAVEL_TIME].margin_of_error_expmech
        )
        args["travel_time_summary_table"] = render_summary(
            report[const.TRAVEL_TIME].quartiles
        )

    if const.JUMP_LENGTH not in dpmreport.analysis_exclusion:
        args["jump_length_eps"] = render_eps(report[const.JUMP_LENGTH].privacy_budget)
        args["jump_length_moe"] = fmt_moe(
            report[const.JUMP_LENGTH].margin_of_error_laplace
        )

        args["jump_length_hist_info"] = render_user_input_info(
            dpmreport.max_jump_length, dpmreport.bin_range_jump_length
        )
        args["jump_length_hist"] = render_jump_length_hist(report[const.JUMP_LENGTH])
        args["jump_length_moe_info"] = render_moe_info(
            report[const.JUMP_LENGTH].margin_of_error_expmech
        )
        args["jump_length_summary_table"] = render_summary(
            report[const.JUMP_LENGTH].quartiles
        )

    template_structure = get_template("od_analysis_segment.html")
    return template_structure.render(args)


def render_benchmark_od_analysis(
    benchmark: "BenchmarkReport",
    top_n_flows: int,
    temp_map_folder: Path,
    output_filename: str,
) -> str:
    report_base = benchmark.report_base.report
    report_alternative = benchmark.report_alternative.report
    tessellation = benchmark.report_base.tessellation
    args: dict = {}
    template_measures = get_template("similarity_measures.html")

    args[
        "user_config_info"
    ] = f"User configuration: display max. top {top_n_flows} OD connections on map"
    args["output_filename"] = output_filename

    if const.OD_FLOWS not in benchmark.analysis_exclusion:
        args["od_eps"] = (
            render_eps(report_base[const.OD_FLOWS].privacy_budget),
            render_eps(report_alternative[const.OD_FLOWS].privacy_budget),
        )
        args["od_moe"] = (
            fmt_moe(report_base[const.OD_FLOWS].margin_of_error_laplace),
            fmt_moe(report_base[const.OD_FLOWS].margin_of_error_laplace),
        )

        args["od_legend"] = render_benchmark_origin_destination_flows(
            report_base[const.OD_FLOWS],
            report_alternative[const.OD_FLOWS],
            tessellation,
            top_n_flows,
            temp_map_folder,
        )

        quartiles_base = report_base[const.OD_FLOWS].quartiles.round()
        quartiles_alternative = report_alternative[const.OD_FLOWS].quartiles.round()
        args["flows_summary_table"] = render_benchmark_summary(
            quartiles_base, quartiles_alternative, target_type=int
        )
        args["flows_cumsum_linechart"] = render_flows_cumsum(
            report_base[const.OD_FLOWS],
            report_alternative[const.OD_FLOWS],
        )
        args["most_freq_flows_ranking"] = render_most_freq_flows_ranking_benchmark(
            report_base[const.OD_FLOWS],
            report_alternative[const.OD_FLOWS],
            tessellation,
        )
        args["flows_measure"] = template_measures.render(
            all_available_measures(const.OD_FLOWS, benchmark)
        )

        args["flows_summary_measure"] = template_measures.render(
            all_available_measures(const.OD_FLOWS_QUARTILES, benchmark)
        )

        args["od_flows_ranking_measure"] = template_measures.render(
            {
                **all_available_measures(const.OD_FLOWS_RANKING, benchmark),
                **{"top_n_object": "flows"},
            }
        )

    if const.TRAVEL_TIME not in benchmark.analysis_exclusion:
        args["travel_time_eps"] = (
            render_eps(report_base[const.TRAVEL_TIME].privacy_budget),
            render_eps(report_alternative[const.TRAVEL_TIME].privacy_budget),
        )
        args["travel_time_moe"] = (
            fmt_moe(report_base[const.TRAVEL_TIME].margin_of_error_laplace),
            fmt_moe(report_alternative[const.TRAVEL_TIME].margin_of_error_laplace),
        )

        args["travel_time_hist_info"] = render_user_input_info(
            benchmark.report_base.max_travel_time,
            benchmark.report_base.bin_range_travel_time,
        )
        args["travel_time_hist"] = render_travel_time_hist(
            report_base[const.TRAVEL_TIME], report_alternative[const.TRAVEL_TIME]
        )
        args["travel_time_moe_info"] = render_moe_info(
            report_base[const.TRAVEL_TIME].margin_of_error_expmech
        )
        args["travel_time_summary_table"] = render_benchmark_summary(
            report_base[const.TRAVEL_TIME].quartiles,
            report_alternative[const.TRAVEL_TIME].quartiles,
            target_type=float,
        )
        args["travel_time_measure"] = template_measures.render(
            all_available_measures(const.TRAVEL_TIME, benchmark)
        )
        args["travel_time_summary_measure"] = template_measures.render(
            all_available_measures(const.TRAVEL_TIME_QUARTILES, benchmark)
        )

    if const.JUMP_LENGTH not in benchmark.analysis_exclusion:
        args["jump_length_eps"] = (
            render_eps(report_base[const.JUMP_LENGTH].privacy_budget),
            render_eps(report_alternative[const.JUMP_LENGTH].privacy_budget),
        )
        args["jump_length_moe"] = (
            fmt_moe(report_base[const.JUMP_LENGTH].margin_of_error_laplace),
            fmt_moe(report_alternative[const.JUMP_LENGTH].margin_of_error_laplace),
        )

        args["jump_length_hist_info"] = render_user_input_info(
            benchmark.report_base.max_jump_length,
            benchmark.report_base.bin_range_jump_length,
        )
        args["jump_length_hist"] = render_jump_length_hist(
            report_base[const.JUMP_LENGTH], report_alternative[const.JUMP_LENGTH]
        )
        args["jump_length_moe_info"] = render_moe_info(
            report_base[const.JUMP_LENGTH].margin_of_error_expmech
        )
        args["jump_length_summary_table"] = render_benchmark_summary(
            report_base[const.JUMP_LENGTH].quartiles,
            report_alternative[const.JUMP_LENGTH].quartiles,
            target_type=float,
        )
        args["jump_length_measure"] = template_measures.render(
            all_available_measures(const.JUMP_LENGTH, benchmark)
        )
        args["jump_length_summary_measure"] = template_measures.render(
            all_available_measures(const.JUMP_LENGTH_QUARTILES, benchmark)
        )

    template_structure = get_template("od_analysis_segment_benchmark.html")
    return template_structure.render(args)


def render_origin_destination_flows(
    od_flows: DfSection,
    tessellation: GeoDataFrame,
    top_n_flows: int,
    threshold: float,
    temp_map_folder: Path,
) -> str:

    data = od_flows.data.copy()
    moe_deviation = od_flows.margin_of_error_laplace / data[const.FLOW]
    data.loc[moe_deviation > threshold, const.FLOW] = None

    # create intra_tile basemap
    intra_tile_flows = data[data[const.ORIGIN] == data[const.DESTINATION]]
    tessellation_intra_flows = pd.merge(
        tessellation,
        intra_tile_flows,
        how="left",
        left_on=const.TILE_ID,
        right_on=const.ORIGIN,
    )

    map, intra_tile_legend = plot.choropleth_map(
        tessellation_intra_flows,
        const.FLOW,
        "intra-tile flows",
        cmap=const.BASE_CMAP,
    )  # get innerflows as color for choropleth

    # create od flows map
    top_n_flows = top_n_flows if top_n_flows <= len(data) else len(data)
    flows = data[
        (data[const.ORIGIN] != data[const.DESTINATION]) & data[const.FLOW].notna()
    ].nlargest(top_n_flows, const.FLOW)
    # only plot lines if there are any flows between tiles
    if not flows[const.FLOW].isnull().all():
        map = plot.flows(map, flows, tessellation)

    map.save(os.path.join(temp_map_folder, "od_map.html"))

    html_legend = v_utils.fig_to_html(intra_tile_legend)
    plt.close()
    return html_legend


def merge_innerflows(
    base_flows: pd.DataFrame, alternative_flows: pd.DataFrame
) -> pd.DataFrame:
    flows = base_flows.merge(
        alternative_flows,
        how="outer",
        on=["origin", "destination"],
        suffixes=["_alt", "_base"],
    )
    flows["flow_base"] = flows["flow_base"] / np.sum(flows["flow_base"])
    flows["flow_alt"] = flows["flow_alt"] / np.sum(flows["flow_alt"])
    flows["flow_base"] = flows["flow_base"].fillna(0)
    flows["flow_alt"] = flows["flow_alt"].fillna(0)
    flows["deviation"] = flows.apply(
        lambda x: symmetric_perc_error(
            x["flow_alt"], x["flow_base"], keep_direction=True
        ),
        axis=1,
    )
    flows.drop(["flow_alt", "flow_base"], axis=1, inplace=True)
    return flows


def render_benchmark_origin_destination_flows(
    od_flows_base: DfSection,
    od_flows_alternative: DfSection,
    tessellation: GeoDataFrame,
    top_n_flows: int,
    temp_map_folder: Path,
) -> str:
    data_base = od_flows_base.data.copy()
    data_alternative = od_flows_alternative.data.copy()

    top_n_flows_base = top_n_flows if top_n_flows <= len(data_base) else len(data_base)
    top_n_flows_alternative = (
        top_n_flows if top_n_flows <= len(data_alternative) else len(data_alternative)
    )

    innerflows = merge_innerflows(
        data_alternative[data_alternative.origin == data_alternative.destination],
        data_base[data_base.origin == data_base.destination],
    )

    tessellation_intra_flows = pd.merge(
        tessellation,
        innerflows,
        how="left",
        left_on=const.TILE_ID,
        right_on="origin",
    )

    tessellation_intra_flows.deviation.fillna(0, inplace=True)

    map, intra_tile_legend = plot.choropleth_map(
        tessellation_intra_flows,
        "deviation",
        "deviation from base intra-tile flows",
        layer_name="Intra-tile flows",
        is_cmap_diverging=True,
        min_scale=-2,
        max_scale=2,
    )  # get innerflows as color for choropleth

    flows_base = data_base[
        (data_base[const.ORIGIN] != data_base[const.DESTINATION])
        & data_base[const.FLOW].notna()
    ].nlargest(top_n_flows_base, const.FLOW)

    flows_alternative = data_alternative[
        (data_alternative[const.ORIGIN] != data_alternative[const.DESTINATION])
        & data_alternative[const.FLOW].notna()
    ].nlargest(top_n_flows_alternative, const.FLOW)

    if not flows_base.flow.isnull().all():
        map = plot.flows(
            map,
            flows_base,
            tessellation,
            flow_color=const.DARK_BLUE,
            marker_color=const.LIGHT_BLUE,
            layer_name="OD flows base",
        )

    if not flows_alternative.flow.isnull().all():
        map = plot.flows(
            map,
            flows_alternative,
            tessellation,
            flow_color=const.ORANGE,
            marker_color=const.LIGHT_ORANGE,
            layer_name="OD flows alternative",
        )

    folium.LayerControl(collapsed=False).add_to(map)

    map.save(os.path.join(temp_map_folder, "od_map.html"))

    html_legend = v_utils.fig_to_html(intra_tile_legend)
    plt.close()
    return html_legend


def render_intra_tile_flows(od_flows: DfSection, n_tiles: int) -> str:
    od_sum = od_flows.data["flow"].sum()
    if od_sum == 0:
        intra_tile_flows_perc = 0
        moe_info = 0
    else:
        intra_tile_flows_perc = round(
            od_analysis.get_intra_tile_flows(od_flows.data) / od_sum * 100, 2
        )
        # TODO: is this actually a correct estimation? -> the more noise the more non-intra tile flows are overestimated...
        moe_info = round(n_tiles * od_flows.margin_of_error_laplace / od_sum * 100, 2)
    ci_interval_info = (
        f"(95% confidence interval ± {moe_info} percentage points)"
        if od_flows.margin_of_error_laplace is not None
        else ""
    )

    return f"{intra_tile_flows_perc}% of flows start and end within the same cell {ci_interval_info}."


def render_flows_cumsum(
    od_flows: DfSection,
    od_flows_alternative: Optional[DfSection] = None,
    diagonal: bool = False,
) -> str:
    df_cumsum = od_flows.cumsum
    if od_flows_alternative:
        df_cumsum_alternative = od_flows_alternative.cumsum
    else:
        df_cumsum_alternative = None

    chart = plot.linechart_new(
        data=df_cumsum,
        x="n",
        y="cum_perc",
        data_alternative=df_cumsum_alternative,
        x_axis_label="Number of OD tile pairs",
        y_axis_label="Cumulated sum of flows between OD pairs",
        add_diagonal=diagonal,
    )
    html = v_utils.fig_to_html(chart)
    plt.close(chart)
    return html


def render_most_freq_flows_ranking(
    od_flows: DfSection, tessellation: GeoDataFrame, top_x: int = 10
) -> str:
    topx_flows = od_flows.data.nlargest(top_x, "flow")

    # if no intra-cell flows should be shown
    # topx_flows = od_flows.data[(od_flows.data.origin != od_flows.data.destination)].nlargest(top_x, "flow") # type: ignore

    topx_flows["rank"] = list(range(1, len(topx_flows) + 1))
    topx_flows = topx_flows.merge(
        tessellation[[const.TILE_ID, const.TILE_NAME]],
        how="left",
        left_on=const.ORIGIN,
        right_on=const.TILE_ID,
    )
    topx_flows = topx_flows.merge(
        tessellation[[const.TILE_ID, const.TILE_NAME]],
        how="left",
        left_on=const.DESTINATION,
        right_on=const.TILE_ID,
        suffixes=("_origin", "_destination"),
    )
    labels = (
        topx_flows["rank"].astype(str)
        + ": "
        + topx_flows[f"{const.TILE_NAME}_origin"]
        + " - \n"
        + topx_flows[f"{const.TILE_NAME}_destination"]
    )

    ranking = plot.ranking(
        topx_flows.flow,
        "Number of flows per OD pair",
        y_labels=labels,
        margin_of_error=od_flows.margin_of_error_laplace,
    )
    html_ranking = v_utils.fig_to_html(ranking)
    plt.close(ranking)
    return html_ranking


def render_most_freq_flows_ranking_benchmark(
    od_flows_base: DfSection,
    od_flows_alternative: DfSection,
    tessellation: GeoDataFrame,
    top_x: int = 10,
) -> str:

    topx_flows_base = od_flows_base.data.nlargest(top_x, "flow")
    topx_flows_base["flow"] = (
        topx_flows_base["flow"] / topx_flows_base["flow"].sum() * 100
    )
    topx_flows_alternative = od_flows_alternative.data.nlargest(top_x, "flow")
    topx_flows_alternative["flow"] = (
        topx_flows_alternative["flow"] / topx_flows_alternative["flow"].sum() * 100
    )

    topx_flows_merged = topx_flows_base.merge(
        topx_flows_alternative,
        on=["origin", "destination"],
        suffixes=("_base", "_alternative"),
        how="outer",
    )
    topx_flows_merged.sort_values(by=["flow_base", "flow_alternative"])
    # if no intra-cell flows should be shown
    # topx_flows = od_flows.data[(od_flows.data.origin != od_flows.data.destination)].nlargest(top_x, "flow") # type: ignore

    topx_flows_merged["rank"] = list(range(1, len(topx_flows_merged) + 1))
    topx_flows_merged = topx_flows_merged.merge(
        tessellation[[const.TILE_ID, const.TILE_NAME]],
        how="left",
        left_on="origin",
        right_on=const.TILE_ID,
    )
    topx_flows_merged = topx_flows_merged.merge(
        tessellation[[const.TILE_ID, const.TILE_NAME]],
        how="left",
        left_on="destination",
        right_on=const.TILE_ID,
        suffixes=("_origin", "_destination"),
    )
    labels = (
        topx_flows_merged["rank"].astype(str)
        + ": "
        + topx_flows_merged[f"{const.TILE_NAME}_origin"]
        + " - \n"
        + topx_flows_merged[f"{const.TILE_NAME}_destination"]
    )

    ranking = plot.ranking(
        x=topx_flows_merged.flow_base,
        x_alternative=topx_flows_merged.flow_alternative,
        x_axis_label="percentage of flows per OD pair",
        y_labels=labels,
        margin_of_error=od_flows_base.margin_of_error_laplace,
        margin_of_error_alternative=od_flows_alternative.margin_of_error_laplace,
    )
    html_ranking = v_utils.fig_to_html(ranking)
    plt.close(ranking)
    return html_ranking


def render_travel_time_hist(
    travel_time_hist: TupleSection,
    travel_time_hist_alternative: Optional[TupleSection] = None,
) -> str:
    if travel_time_hist_alternative:
        alternative_data = travel_time_hist_alternative.data
        alternative_moe = travel_time_hist_alternative.margin_of_error_laplace
    else:
        alternative_data = None
        alternative_moe = None
    hist = plot.histogram(
        hist=travel_time_hist.data,
        hist_alternative=alternative_data,
        x_axis_label="travel time (min.)",
        y_axis_label="% of trips",
        x_axis_type=int,
        margin_of_error=travel_time_hist.margin_of_error_laplace,
        margin_of_error_alternative=alternative_moe,
    )
    html_hist = v_utils.fig_to_html(hist)
    plt.close()
    return html_hist


def render_jump_length_hist(
    jump_length_hist: TupleSection,
    jump_length_hist_alternative: Optional[TupleSection] = None,
) -> str:
    if jump_length_hist_alternative:
        alternative_data = jump_length_hist_alternative.data
        alternative_moe = jump_length_hist_alternative.margin_of_error_laplace
    else:
        alternative_data = None
        alternative_moe = None

    hist = plot.histogram(
        hist=jump_length_hist.data,
        hist_alternative=alternative_data,
        x_axis_label="jump length (kilometers)",
        y_axis_label="% of trips",
        x_axis_type=float,
        margin_of_error=jump_length_hist.margin_of_error_laplace,
        margin_of_error_alternative=alternative_moe,
    )
    html_hist = v_utils.fig_to_html(hist)
    plt.close()
    return html_hist



import warnings
from datetime import timedelta
from typing import TYPE_CHECKING, Any, List, Optional, Union

import cv2
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from haversine import Unit, haversine
from scipy.spatial.distance import jensenshannon
from scipy.stats import entropy, kendalltau, wasserstein_distance
from tqdm.auto import tqdm

from dp_mobility_report import constants as const

if TYPE_CHECKING:
    from dp_mobility_report import BenchmarkReport


# catch warning from scipy
def _entropy(pk: np.array, qk: np.array) -> float:
    if (np.sum(pk) == 0) | (np.sum(qk) == 0):
        return np.nan
    else:
        return entropy(pk, qk)


# catch warning from scipy
def _jensenshannon(p: np.array, q: np.array) -> float:
    if (np.sum(p) == 0) | (np.sum(q) == 0):
        return np.nan
    else:
        return jensenshannon(p, q)


def _moving_average(arr: np.array, size: int) -> np.array:
    return np.convolve(arr, np.ones(size), "valid") / size


def _replace_inf_with_max(
    bins: np.array, max_value: Union[int, float, timedelta]
) -> np.array:
    if np.isinf(bins[-1]):
        if isinstance(max_value, timedelta):  # needed for user_time_delta
            max_value = max_value.total_seconds() / 3600
        return np.append(bins[:-1], max_value)
    return bins


def earth_movers_distance1D(
    u_hist: tuple, v_hist: tuple, u_max: Union[int, float], v_max: Union[int, float]
) -> float:
    # in emd terms, hist_bins(x-axis) are values and hist_values(y-axis) are weights

    # bins greater than defined hist_max values are summarized as a "greater" bin, named "inf"
    # to compute the emd, the actual max value is needed
    u_hist_bins = _replace_inf_with_max(u_hist[1], u_max)
    v_hist_bins = _replace_inf_with_max(v_hist[1], v_max)

    # distinguish between single int bins and ranges
    if len(u_hist[0]) == len(u_hist_bins):
        u_values = u_hist_bins
        v_values = v_hist_bins
    else:  # computes moving average if hist buckets within a range
        u_values = _moving_average(u_hist_bins, 2)
        v_values = _moving_average(v_hist_bins, 2)

    u_weights = u_hist[0]
    v_weights = v_hist[0]
    if (sum(u_weights) == 0) | (sum(v_weights) == 0):
        return np.nan
    return wasserstein_distance(u_values, v_values, u_weights, v_weights)


# keep_direction: whether to use absolute difference or keep direction of difference
def symmetric_perc_error(
    alternative: Optional[Union[int, float]],
    base: Optional[Union[int, float]],
    keep_direction: bool = False,
) -> float:
    if (alternative is None) or (base is None):
        return None
    if abs(base) + abs(alternative) == 0:
        return 0

    diff_alt_base = alternative - base
    if not keep_direction:
        diff_alt_base = np.abs(diff_alt_base)

    return diff_alt_base / ((abs(base) + abs(alternative)) / 2)


def symmetric_mape(
    alternative: Union[pd.Series, np.array],
    base: Union[pd.Series, np.array],
) -> float:
    n = len(base)
    return (
        1 / n * np.sum([symmetric_perc_error(a, b) for a, b in zip(alternative, base)])
    )


def all_perc_errors(alt_dict: dict, base_dict: dict) -> dict:
    re = {}
    for key in base_dict:
        re[key] = symmetric_perc_error(alt_dict[key], base_dict[key])
    return re


# earth movers distance
def _compute_cost_matrix(tessellation: GeoDataFrame) -> np.array:
    tile_centroids = (
        tessellation.set_index(const.TILE_ID).to_crs(3395).centroid.to_crs(4326)
    )
    tile_coords = list(zip(tile_centroids.y, tile_centroids.x))

    # get all potential combinations between all points from sig1 and sig2
    grid = np.meshgrid(range(0, len(tile_coords)), range(0, len(tile_coords)))
    tile_combinations = np.array([grid[0].flatten(), grid[1].flatten()])

    # create an empty cost matrix with the length of all possible combinations
    cost_matrix = np.empty(
        tile_combinations.shape[1], dtype=np.float32
    )  # float32 needed as input for cv2.emd!

    # compute haversine distance for all possible combinations
    for column in range(0, tile_combinations.shape[1]):
        tile_1 = tile_combinations[0, column]
        tile_2 = tile_combinations[1, column]
        cost_matrix[column] = haversine(
            tile_coords[tile_1], tile_coords[tile_2], unit=Unit.METERS
        )

    # reshape array to matrix
    return np.reshape(cost_matrix, (len(tile_coords), len(tile_coords)))


def earth_movers_distance(
    arr_alt: np.array, arr_base: np.array, cost_matrix: np.array
) -> float:  # based on haversine distance

    # set values to 1 (as before) or stay with nan, as now? especially for time windows?
    if (all(arr_alt == 0)) | (all(arr_base == 0)):
        return np.nan
    arr_base = (arr_base / arr_base.sum() * 100).round(2)
    sig_true = arr_base.astype(np.float32)

    # normalize input and assign needed type for cv2
    arr_alt = (arr_alt / arr_alt.sum() * 100).round(2)
    sig_estimate = arr_alt.astype(np.float32)

    emd_dist, _, _ = cv2.EMD(
        sig_true, sig_estimate, distType=cv2.DIST_USER, cost=cost_matrix
    )
    return emd_dist


def top_n_coverage(top_n_tiles_base: list, top_n_tiles_alt: list) -> float:
    top_n_base = len(top_n_tiles_base)
    if top_n_base == 0:
        return 0
    top_n_alternative = len([x for x in top_n_tiles_alt if x in top_n_tiles_base])
    return top_n_alternative / top_n_base


def compute_similarity_measures(
    analysis_exclusion: list,
    report_alternative: dict,
    report_base: dict,
    tessellation: Optional[GeoDataFrame],
    top_n_ranking: List[int],
    disable_progress_bar: bool,
) -> Any:
    smape_dict: dict = {}
    kld_dict: dict = {}
    jsd_dict: dict = {}
    emd_dict: dict = {}
    kendall_dict: dict = {}
    top_n_coverage_dict: dict = {}
    cost_matrix = None

    disable_emd = (
        len(tessellation) > const.DISABLE_EMD_THRESHOLD
        if (tessellation is not None)
        else True
    )

    if disable_emd and (
        (const.VISITS_PER_TILE not in analysis_exclusion)
        or (const.VISITS_PER_TIME_TILE not in analysis_exclusion)
    ):
        warnings.warn(
            "EMD computation disables for spatial analysis due to too many tiles in tessellation (results in long computation time)."
        )

    with tqdm(  # progress bar
        total=15, desc="Compute similarity measures:", disable=disable_progress_bar
    ) as pbar:

        # Statistics
        if const.DS_STATISTICS not in analysis_exclusion:
            smape_dict = dict(
                **smape_dict,
                **all_perc_errors(
                    report_alternative[const.DS_STATISTICS].data,
                    report_base[const.DS_STATISTICS].data,
                ),
            )
        pbar.update()

        # Missing values
        if const.MISSING_VALUES not in analysis_exclusion:
            smape_dict = dict(
                **smape_dict,
                **all_perc_errors(
                    report_alternative[const.MISSING_VALUES].data,
                    report_base[const.MISSING_VALUES].data,
                ),
            )
        pbar.update()
        # Temporal distributions
        if const.TRIPS_OVER_TIME not in analysis_exclusion:
            trips_over_time = report_alternative[const.TRIPS_OVER_TIME].data.merge(
                report_base[const.TRIPS_OVER_TIME].data,
                how="outer",
                on="datetime",
                suffixes=("_alternative", "_base"),
            )

            trips_over_time.fillna(0, inplace=True)
            kld_dict[const.TRIPS_OVER_TIME] = _entropy(
                pk=trips_over_time.trips_alternative, qk=trips_over_time.trips_base
            )
            jsd_dict[const.TRIPS_OVER_TIME] = _jensenshannon(
                p=trips_over_time.trips_alternative, q=trips_over_time.trips_base
            )
            smape_dict[const.TRIPS_OVER_TIME] = symmetric_mape(
                alternative=trips_over_time.trips_alternative,
                base=trips_over_time.trip_count_base,
            )
        pbar.update()

        if const.TRIPS_PER_WEEKDAY not in analysis_exclusion:
            trips_per_weekday = pd.concat(
                [
                    report_alternative[const.TRIPS_PER_WEEKDAY].data,
                    report_base[const.TRIPS_PER_WEEKDAY].data,
                ],
                join="outer",
                axis=1,
            )
            trips_per_weekday.fillna(0, inplace=True)
            kld_dict[const.TRIPS_PER_WEEKDAY] = _entropy(
                pk=trips_per_weekday.iloc[:, 0], qk=trips_per_weekday.iloc[:, 1]
            )
            jsd_dict[const.TRIPS_PER_WEEKDAY] = _jensenshannon(
                p=trips_per_weekday.iloc[:, 0], q=trips_per_weekday.iloc[:, 1]
            )
            smape_dict[const.TRIPS_PER_WEEKDAY] = symmetric_mape(
                alternative=trips_per_weekday.iloc[:, 0],
                base=trips_per_weekday.iloc[:, 1],
            )
        pbar.update()

        if const.TRIPS_PER_HOUR not in analysis_exclusion:
            trips_per_hour = report_alternative[const.TRIPS_PER_HOUR].data.merge(
                report_base[const.TRIPS_PER_HOUR].data,
                how="outer",
                on=["hour", const.TIME_CATEGORY],
                suffixes=("_alternative", "_base"),
            )
            trips_per_hour.fillna(0, inplace=True)
            kld_dict[const.TRIPS_PER_HOUR] = _entropy(
                pk=trips_per_hour.perc_alternative, qk=trips_per_hour.perc_base
            )
            jsd_dict[const.TRIPS_PER_HOUR] = _jensenshannon(
                p=trips_per_hour.perc_alternative, q=trips_per_hour.perc_base
            )
            smape_dict[const.TRIPS_PER_HOUR] = symmetric_mape(
                alternative=trips_per_hour.perc_alternative,
                base=trips_per_hour.perc_base,
            )
        pbar.update()

        # Spatial distribution
        if const.VISITS_PER_TILE not in analysis_exclusion:
            tessellation = tessellation.sort_values(by=const.TILE_ID)
            visits_per_tile = report_alternative[const.VISITS_PER_TILE].data.merge(
                report_base[const.VISITS_PER_TILE].data,
                how="outer",
                on="tile_id",
                suffixes=("_alternative", "_base"),
                sort=True,  # sort according to tessellation for cost_matrix
            )
            visits_per_tile.fillna(0, inplace=True)

            rel_counts_alternative = (
                visits_per_tile.visits_alternative
                / visits_per_tile.visits_alternative.sum()
            )
            rel_counts_base = (
                visits_per_tile.visits_base / visits_per_tile.visits_base.sum()
            )
            kld_dict[const.VISITS_PER_TILE] = _entropy(
                pk=rel_counts_alternative, qk=rel_counts_base
            )
            jsd_dict[const.VISITS_PER_TILE] = _jensenshannon(
                p=rel_counts_alternative, q=rel_counts_base
            )
            smape_dict[const.VISITS_PER_TILE] = symmetric_mape(
                alternative=rel_counts_alternative, base=rel_counts_base
            )

            most_freq_base = report_base[const.VISITS_PER_TILE].data.sort_values(
                by=["visits"], ascending=False, ignore_index=True
            )
            most_freq_alternative = report_alternative[
                const.VISITS_PER_TILE
            ].data.sort_values(by=["visits"], ascending=False, ignore_index=True)

            # mute kendall_tau warning by replacing str tile_ids with ints
            tile_id_to_index = {
                tile_id: i
                for tile_id, i in zip(
                    tessellation[const.TILE_ID], range(0, len(tessellation))
                )
            }
            kendall_dict[const.VISITS_PER_TILE_RANKING] = []
            top_n_coverage_dict[const.VISITS_PER_TILE_RANKING] = []
            for top_n in top_n_ranking:
                kt_temp, _ = kendalltau(
                    [
                        tile_id_to_index[tile_id]
                        for tile_id in most_freq_base.truncate(after=top_n)[
                            const.TILE_ID
                        ]
                    ],
                    [
                        tile_id_to_index[tile_id]
                        for tile_id in most_freq_alternative.truncate(after=top_n)[
                            const.TILE_ID
                        ]
                    ],
                )
                kendall_dict[const.VISITS_PER_TILE_RANKING] += [kt_temp]
                top_n_coverage_dict[const.VISITS_PER_TILE_RANKING] += [
                    top_n_coverage(
                        list(most_freq_base.truncate(after=top_n)[const.TILE_ID]),
                        list(
                            most_freq_alternative.truncate(after=top_n)[const.TILE_ID]
                        ),
                    )
                ]

            if not disable_emd:
                # create custom cost matrix with distances between all tiles
                cost_matrix = _compute_cost_matrix(tessellation)

                emd_dict[const.VISITS_PER_TILE] = earth_movers_distance(
                    arr_alt=visits_per_tile.visits_alternative.to_numpy(),
                    arr_base=visits_per_tile.visits_base.to_numpy(),
                    cost_matrix=cost_matrix,
                )

            # Outliers
            smape_dict[const.VISITS_PER_TILE_OUTLIERS] = symmetric_perc_error(
                report_alternative[const.VISITS_PER_TILE].n_outliers,
                report_base[const.VISITS_PER_TILE].n_outliers,
            )

            # Quartiles
            smape_dict[const.VISITS_PER_TILE_QUARTILES] = symmetric_mape(
                alternative=report_alternative[const.VISITS_PER_TILE].quartiles,
                base=report_base[const.VISITS_PER_TILE].quartiles,
            )
        pbar.update()

        # Spatio-temporal distributions
        if const.VISITS_PER_TIME_TILE not in analysis_exclusion:

            counts_timew_alternative = (
                report_alternative[const.VISITS_PER_TIME_TILE]
                .data[
                    report_alternative[const.VISITS_PER_TIME_TILE].data.index != "None"
                ]
                .unstack()
            )
            counts_timew_base = (
                report_base[const.VISITS_PER_TIME_TILE]
                .data[report_base[const.VISITS_PER_TIME_TILE].data.index != "None"]
                .unstack()
            )

            indices = np.unique(
                np.append(
                    counts_timew_alternative.index.values,
                    counts_timew_base.index.values,
                )
            )
            counts_timew_alternative = counts_timew_alternative.reindex(index=indices)
            counts_timew_alternative.fillna(0, inplace=True)

            counts_timew_base = counts_timew_base.reindex(index=indices)
            counts_timew_base.fillna(0, inplace=True)

            rel_counts_timew_alternative = (
                counts_timew_alternative / counts_timew_alternative.sum()
            )
            rel_counts_timew_base = counts_timew_base / counts_timew_base.sum()

            kld_dict[const.VISITS_PER_TIME_TILE] = _entropy(
                pk=rel_counts_timew_alternative.to_numpy().flatten(),
                qk=rel_counts_timew_base.to_numpy().flatten(),
            )
            jsd_dict[const.VISITS_PER_TIME_TILE] = _jensenshannon(
                p=rel_counts_timew_alternative.to_numpy().flatten(),
                q=rel_counts_timew_base.to_numpy().flatten(),
            )
            smape_dict[const.VISITS_PER_TIME_TILE] = symmetric_mape(
                alternative=rel_counts_timew_alternative.to_numpy().flatten(),
                base=rel_counts_timew_base.to_numpy().flatten(),
            )
            if not disable_emd:
                if cost_matrix is None:
                    tessellation = tessellation.sort_values(by=const.TILE_ID)
                    cost_matrix = _compute_cost_matrix(tessellation)

                visits_per_time_tile_emd = []
                for time_window in report_base[const.VISITS_PER_TIME_TILE].data.columns:
                    tw_base = report_base[const.VISITS_PER_TIME_TILE].data[time_window]
                    tw_base = tw_base / tw_base.sum()
                    # if time window not in proposal report, add time windows with count zero
                    if (
                        time_window
                        not in report_alternative[
                            const.VISITS_PER_TIME_TILE
                        ].data.columns
                    ):
                        tw_alternative = tw_base.copy()
                        tw_alternative[:] = 0
                    else:
                        tw_alternative = report_alternative[
                            const.VISITS_PER_TIME_TILE
                        ].data[time_window]
                        tw_alternative = tw_alternative / tw_alternative.sum()
                    tw = pd.merge(
                        tw_alternative,
                        tw_base,
                        how="outer",
                        right_index=True,
                        left_index=True,
                        suffixes=("_alternative", "_base"),
                        sort=True,  # sort according to tile_id for cost_matrix
                    )

                    tw = tw[
                        tw.notna().sum(axis=1) > 0
                    ]  # remove instances where both are NaN
                    tw.fillna(0, inplace=True)
                    visits_per_time_tile_emd.append(
                        earth_movers_distance(
                            arr_alt=tw.iloc[:, 0].to_numpy(),
                            arr_base=tw.iloc[:, 1].to_numpy(),
                            cost_matrix=cost_matrix,
                        )
                    )
                emd_dict[const.VISITS_PER_TIME_TILE] = np.mean(visits_per_time_tile_emd)

        pbar.update()

        # Origin-Destination
        if const.OD_FLOWS not in analysis_exclusion:
            all_od_combinations = pd.concat(
                [
                    report_alternative[const.OD_FLOWS].data[["origin", "destination"]],
                    report_base[const.OD_FLOWS].data[["origin", "destination"]],
                ]
            ).drop_duplicates()
            all_od_combinations["flow"] = 0

            od_flows_alternative = (
                pd.concat(
                    [report_alternative[const.OD_FLOWS].data, all_od_combinations]
                )
                .drop_duplicates(["origin", "destination"], keep="first")
                .sort_values(["origin", "destination"])
                .flow
            )
            od_flows_base = (
                pd.concat([report_base[const.OD_FLOWS].data, all_od_combinations])
                .drop_duplicates(["origin", "destination"], keep="first")
                .sort_values(["origin", "destination"])
                .flow
            )

            rel_alternative = od_flows_alternative / (od_flows_alternative.sum())
            rel_base = od_flows_base / od_flows_base.sum()

            kld_dict[const.OD_FLOWS] = _entropy(
                pk=rel_alternative.to_numpy(), qk=rel_base.to_numpy()
            )
            jsd_dict[const.OD_FLOWS] = _jensenshannon(
                p=rel_alternative.to_numpy(), q=rel_base.to_numpy()
            )
            smape_dict[const.OD_FLOWS] = symmetric_mape(
                alternative=rel_alternative.to_numpy(), base=rel_base.to_numpy()
            )

            kendall_dict[const.OD_FLOWS_RANKING] = []
            top_n_coverage_dict[const.OD_FLOWS_RANKING] = []
            for top_n in top_n_ranking:
                kt_temp, _ = kendalltau(
                    list(rel_alternative.sort_values(ascending=False).index.values)[
                        :top_n
                    ],
                    list(rel_base.sort_values(ascending=False).index.values)[:top_n],
                )
                kendall_dict[const.OD_FLOWS_RANKING] += [kt_temp]
                top_n_coverage_dict[const.OD_FLOWS_RANKING] += [
                    top_n_coverage(
                        list(
                            od_flows_base.sort_values(ascending=False)
                            .iloc[0:top_n]
                            .index
                        ),
                        list(
                            od_flows_alternative.sort_values(ascending=False)
                            .iloc[0:top_n]
                            .index
                        ),
                    )
                ]

            # Quartiles
            smape_dict[const.OD_FLOWS_QUARTILES] = symmetric_mape(
                alternative=report_alternative[const.OD_FLOWS].quartiles,
                base=report_base[const.OD_FLOWS].quartiles,
            )
        pbar.update()

        if const.TRAVEL_TIME not in analysis_exclusion:
            kld_dict[const.TRAVEL_TIME] = _entropy(
                pk=report_alternative[const.TRAVEL_TIME].data[0],
                qk=report_base[const.TRAVEL_TIME].data[0],
            )
            jsd_dict[const.TRAVEL_TIME] = _jensenshannon(
                p=report_alternative[const.TRAVEL_TIME].data[0],
                q=report_base[const.TRAVEL_TIME].data[0],
            )
            smape_dict[const.TRAVEL_TIME] = symmetric_mape(
                alternative=report_alternative[const.TRAVEL_TIME].data[0],
                base=report_base[const.TRAVEL_TIME].data[0],
            )
            emd_dict[const.TRAVEL_TIME] = earth_movers_distance1D(
                report_alternative[const.TRAVEL_TIME].data,
                report_base[const.TRAVEL_TIME].data,
                report_alternative[const.TRAVEL_TIME].quartiles["max"],
                report_base[const.TRAVEL_TIME].quartiles["max"],
            )
            # Quartiles
            smape_dict[const.TRAVEL_TIME_QUARTILES] = symmetric_mape(
                alternative=report_alternative[const.TRAVEL_TIME].quartiles,
                base=report_base[const.TRAVEL_TIME].quartiles,
            )
        pbar.update()

        if const.JUMP_LENGTH not in analysis_exclusion:
            kld_dict[const.JUMP_LENGTH] = _entropy(
                pk=report_alternative[const.JUMP_LENGTH].data[0],
                qk=report_base[const.JUMP_LENGTH].data[0],
            )
            jsd_dict[const.JUMP_LENGTH] = _jensenshannon(
                p=report_alternative[const.JUMP_LENGTH].data[0],
                q=report_base[const.JUMP_LENGTH].data[0],
            )
            smape_dict[const.JUMP_LENGTH] = symmetric_mape(
                alternative=report_alternative[const.JUMP_LENGTH].data[0],
                base=report_base[const.JUMP_LENGTH].data[0],
            )
            emd_dict[const.JUMP_LENGTH] = earth_movers_distance1D(
                report_alternative[const.JUMP_LENGTH].data,
                report_base[const.JUMP_LENGTH].data,
                report_alternative[const.JUMP_LENGTH].quartiles["max"],
                report_base[const.JUMP_LENGTH].quartiles["max"],
            )
            # Quartiles
            smape_dict[const.JUMP_LENGTH_QUARTILES] = symmetric_mape(
                alternative=report_alternative[const.JUMP_LENGTH].quartiles,
                base=report_base[const.JUMP_LENGTH].quartiles,
            )
        pbar.update()

        # User
        # TODO bin sizes do not align for kld, jsd, smape
        if const.TRIPS_PER_USER not in analysis_exclusion:

            emd_dict[const.TRIPS_PER_USER] = earth_movers_distance1D(
                report_alternative[const.TRIPS_PER_USER].data,
                report_base[const.TRIPS_PER_USER].data,
                report_alternative[const.TRIPS_PER_USER].quartiles["max"],
                report_base[const.TRIPS_PER_USER].quartiles["max"],
            )

            # Quartiles
            smape_dict[const.TRIPS_PER_USER_QUARTILES] = symmetric_mape(
                alternative=report_alternative[const.TRIPS_PER_USER].quartiles,
                base=report_base[const.TRIPS_PER_USER].quartiles,
            )
        pbar.update()

        if const.USER_TIME_DELTA not in analysis_exclusion:
            if (
                report_alternative[const.USER_TIME_DELTA] is None
            ):  # if each user only has one trip then `USER_TIME_DELTA` is None
                smape_dict[const.USER_TIME_DELTA_QUARTILES] = None
            else:
                kld_dict[const.USER_TIME_DELTA] = _entropy(
                    pk=report_alternative[const.USER_TIME_DELTA].data[0],
                    qk=report_base[const.USER_TIME_DELTA].data[0],
                )
                jsd_dict[const.USER_TIME_DELTA] = _jensenshannon(
                    p=report_alternative[const.USER_TIME_DELTA].data[0],
                    q=report_base[const.USER_TIME_DELTA].data[0],
                )
                smape_dict[const.USER_TIME_DELTA] = symmetric_mape(
                    alternative=report_alternative[const.USER_TIME_DELTA].data[0],
                    base=report_base[const.USER_TIME_DELTA].data[0],
                )
                emd_dict[const.USER_TIME_DELTA] = earth_movers_distance1D(
                    report_alternative[const.USER_TIME_DELTA].data,
                    report_base[const.USER_TIME_DELTA].data,
                    report_alternative[const.USER_TIME_DELTA].quartiles["max"],
                    report_base[const.USER_TIME_DELTA].quartiles["max"],
                )
                smape_dict[const.USER_TIME_DELTA_QUARTILES] = symmetric_mape(
                    alternative=report_alternative[
                        const.USER_TIME_DELTA
                    ].quartiles.apply(lambda x: x.total_seconds() / 3600),
                    base=report_base[const.USER_TIME_DELTA].quartiles.apply(
                        lambda x: x.total_seconds() / 3600
                    ),
                )
        pbar.update()

        if const.RADIUS_OF_GYRATION not in analysis_exclusion:

            kld_dict[const.RADIUS_OF_GYRATION] = _entropy(
                pk=report_alternative[const.RADIUS_OF_GYRATION].data[0],
                qk=report_base[const.RADIUS_OF_GYRATION].data[0],
            )
            jsd_dict[const.RADIUS_OF_GYRATION] = _jensenshannon(
                p=report_alternative[const.RADIUS_OF_GYRATION].data[0],
                q=report_base[const.RADIUS_OF_GYRATION].data[0],
            )
            emd_dict[const.RADIUS_OF_GYRATION] = earth_movers_distance1D(
                report_alternative[const.RADIUS_OF_GYRATION].data,
                report_base[const.RADIUS_OF_GYRATION].data,
                report_alternative[const.RADIUS_OF_GYRATION].quartiles["max"],
                report_base[const.RADIUS_OF_GYRATION].quartiles["max"],
            )
            smape_dict[const.RADIUS_OF_GYRATION] = symmetric_mape(
                alternative=report_alternative[const.RADIUS_OF_GYRATION].data[0],
                base=report_base[const.RADIUS_OF_GYRATION].data[0],
            )
            # Quartiles
            smape_dict[const.RADIUS_OF_GYRATION_QUARTILES] = symmetric_mape(
                alternative=report_alternative[const.RADIUS_OF_GYRATION].quartiles,
                base=report_base[const.RADIUS_OF_GYRATION].quartiles,
            )
        pbar.update()

        if const.USER_TILE_COUNT not in analysis_exclusion:
            kld_dict[const.USER_TILE_COUNT] = _entropy(
                pk=report_alternative[const.USER_TILE_COUNT].data[0],
                qk=report_base[const.USER_TILE_COUNT].data[0],
            )
            jsd_dict[const.USER_TILE_COUNT] = _jensenshannon(
                p=report_alternative[const.USER_TILE_COUNT].data[0],
                q=report_base[const.USER_TILE_COUNT].data[0],
            )
            emd_dict[const.USER_TILE_COUNT] = earth_movers_distance1D(
                report_alternative[const.USER_TILE_COUNT].data,
                report_base[const.USER_TILE_COUNT].data,
                report_alternative[const.USER_TILE_COUNT].quartiles["max"],
                report_base[const.USER_TILE_COUNT].quartiles["max"],
            )
            smape_dict[const.USER_TILE_COUNT] = symmetric_mape(
                alternative=report_alternative[const.USER_TILE_COUNT].data[0],
                base=report_base[const.USER_TILE_COUNT].data[0],
            )

            smape_dict[const.USER_TILE_COUNT_QUARTILES] = symmetric_mape(
                alternative=report_alternative[const.USER_TILE_COUNT].quartiles,
                base=report_base[const.USER_TILE_COUNT].quartiles,
            )
        pbar.update()

        if const.MOBILITY_ENTROPY not in analysis_exclusion:
            kld_dict[const.MOBILITY_ENTROPY] = _entropy(
                pk=report_alternative[const.MOBILITY_ENTROPY].data[0],
                qk=report_base[const.MOBILITY_ENTROPY].data[0],
            )
            jsd_dict[const.MOBILITY_ENTROPY] = _jensenshannon(
                p=report_alternative[const.MOBILITY_ENTROPY].data[0],
                q=report_base[const.MOBILITY_ENTROPY].data[0],
            )
            emd_dict[const.MOBILITY_ENTROPY] = earth_movers_distance1D(
                report_alternative[const.MOBILITY_ENTROPY].data,
                report_base[const.MOBILITY_ENTROPY].data,
                report_alternative[const.MOBILITY_ENTROPY].quartiles["max"],
                report_base[const.MOBILITY_ENTROPY].quartiles["max"],
            )
            smape_dict[const.MOBILITY_ENTROPY] = symmetric_mape(
                alternative=report_alternative[const.MOBILITY_ENTROPY].data[0],
                base=report_base[const.MOBILITY_ENTROPY].data[0],
            )
            # Quartiles
            smape_dict[const.MOBILITY_ENTROPY_QUARTILES] = symmetric_mape(
                alternative=report_alternative[const.MOBILITY_ENTROPY].quartiles,
                base=report_base[const.MOBILITY_ENTROPY].quartiles,
            )
        pbar.update()

    return smape_dict, kld_dict, jsd_dict, emd_dict, kendall_dict, top_n_coverage_dict


def get_selected_measures(benchmarkreport: "BenchmarkReport") -> dict:
    similarity_measures = {}

    for analysis in benchmarkreport.measure_selection.keys():
        selected_measure = benchmarkreport.measure_selection[analysis]
        try:
            if selected_measure == const.SMAPE:
                if analysis == const.DS_STATISTICS:
                    for element in const.DS_STATISTICS_ELEMENTS:
                        similarity_measures[element] = benchmarkreport.smape[element]
                elif analysis == const.MISSING_VALUES:
                    for element in const.MISSING_VALUES_ELEMENTS:
                        similarity_measures[element] = benchmarkreport.smape[element]
                else:
                    similarity_measures[analysis] = benchmarkreport.smape[analysis]
            elif selected_measure == const.KLD:
                similarity_measures[analysis] = benchmarkreport.kld[analysis]
            elif selected_measure == const.JSD:
                similarity_measures[analysis] = benchmarkreport.jsd[analysis]
            elif selected_measure == const.EMD:
                similarity_measures[analysis] = benchmarkreport.emd[analysis]
            elif selected_measure == const.KT:
                similarity_measures[analysis] = benchmarkreport.kt[analysis]
        except KeyError:
            warnings.warn(
                f"The selected measure {selected_measure} for {analysis} cannot be computed. Value for {analysis} in `self.similarity_measures` will be set to `None`."
            )
            similarity_measures[analysis] = None

    return similarity_measures



import os
import warnings
from pathlib import Path
from shutil import rmtree
from typing import List, Optional, Union

import numpy as np
from geopandas import GeoDataFrame
from pandas import DataFrame

from dp_mobility_report.benchmark import b_utils, preprocessing
from dp_mobility_report.benchmark.similarity_measures import (
    compute_similarity_measures,
    get_selected_measures,
)
from dp_mobility_report.dpmreport import DpMobilityReport
from dp_mobility_report.report.html.templates import (
    create_html_assets,
    create_maps_folder,
    render_benchmark_html,
)


class BenchmarkReport:
    """Evaluate the similarity of two (differentially private) mobility reports from one or two mobility datasets.
        This can be based on two datasets (``df_base`` and ``df_alternative``) or one dataset (``df_base``) with different privacy settings.
        The arguments ``df``, ``privacy_budget``, ``user_privacy``, ``max_trips_per_user`` and ``budget_split`` can differ for the two datasets set with the according ending ``_base`` and ``_alternative``. The other arguments are the same for both reports.
        For the evaluation, similarity measures (namely the symmetric mean absolute percentage error (SMAPE), Jensen-Shannon divergence (JSD), Kullback-Leibler divergence (KLD), the earth mover's distance (EMD), the Kendall correlation coefficient (KT) and the top n coverage (TOP_N_COV)) are computed to quantify the statistical similarity for each analysis.
        The evaluation, i.e., benchmark report, will be generated as an HTML file, using the ``.to_file()`` method.

    Args:
        df_base: ``DataFrame`` containing the baseline mobility data, see argument ``df`` of ``DpMobilityReport``.
        tessellation: Geopandas ``GeoDataFrame`` containing the tessellation for spatial aggregations. Expected columns: ``tile_id``. If tessellation is not provided in the expected default CRS EPSG:4326 it will automatically be transformed. If no tessellation is provided, all analyses based on the tessellation will automatically be removed.
        df_alternative: ``DataFrame`` containing the alternative mobility data to be compared against the baseline dataset, see argument ``df`` of ``DpMobilityReport``. If ``None``, ``df_base`` is used for both reports.
        privacy_budget_base: privacy_budget for the differentially private base report. Defaults to ``None``, i.e., no privacy guarantee is provided.
        privacy_budget_alternative: privacy_budget for the differentially private alternative report. Defaults to ``None``, i.e., no privacy guarantee is provided.
        user_privacy_base: Whether item-level or user-level privacy is applied for the base report. Defaults to ``True`` (user-level privacy).
        user_privacy_alternative: Whether item-level or user-level privacy is applied for the alternative report. Defaults to ``True`` (user-level privacy).
        max_trips_per_user_base: maximum number of trips a user shall contribute to the data. Dataset will be sampled accordingly. Defaults to ``None``, i.e., all trips included.
        max_trips_per_user_alternative: maximum number of trips a user shall contribute to the data. Dataset will be sampled accordingly. Defaults to ``None``, i.e., all trips included.
        analysis_selection: Select only needed analyses, see argument ``analysis_selection`` of ``DpMobilityReport``.
        analysis_exclusion: Ignored, if ``analysis_selection`` is set! Exclude analyses that are not needed, see argument ````analysis_exclusion```` of ``DpMobilityReport``.
        budget_split_base: ``dict``to customize how much privacy budget is assigned to which analysis. See argument ``budget_split`` of ``DpMobilityReport``.
        budget_split_alternative: ``dict``to customize how much privacy budget is assigned to which analysis. See argument ``budget_split`` of ``DpMobilityReport``.
        timewindows: List of hours as ``int`` that define the timewindows for the spatial analysis for single time windows. Defaults to [2, 6, 10, 14, 18, 22].
        max_travel_time: Upper bound for travel time histogram. Defaults to 120 (mins).
        bin_range_travel_time: The range a single histogram bin spans for travel time (e.g., 5 for 5 min bins). Defaults to 5 (min).
        max_jump_length: Upper bound for jump length histogram. Defaults to 10 (km).
        bin_range_jump_length: The range a single histogram bin spans for jump length (e.g., 1 for 1 km bins). Defaults to 1 (km).
        max_radius_of_gyration: Upper bound for radius of gyration histogram. Defaults to 5 (km).
        bin_range_radius_of_gyration: The range a single histogram bin spans for the radius of gyration (e.g., 1 for 1 km bins). Defaults to 0.5 (km).
        max_user_tile_count: Upper bound for distinct tiles per user histogram. Defaults to 10.
        bin_range_user_tile_count: The range a single histogram bin spans for the distinct tiles per user histogram. Defaults to 1.
        max_user_time_delta:  Upper bound for user time delta histogram. Defaults to 48 (hours).
        bin_range_user_time_delta: The range a single histogram bin spans for user time delta (e.g., 1 for 1 hour bins). Defaults to 4 (hours).
        top_n_ranking: List of 'top n' values that are used to compute the Kendall correlation coefficient and the top n coverage for ranking similarity measures. Values need to be integers > 0. Defaults to ``[10, 50, 100]``.
        measure_selection: Select similarity measure for each analysis that is used for the ``similarity_measures`` property of the ``BenchmarkReport``. If ``None``, the default from ``default_measure_selection()`` will be used.
        disable_progress_bar: Whether progress bars should be shown. Defaults to ``False``.
        seed_sampling: Provide seed for down-sampling of dataset (according to ``max_trips_per_user``) so that the sampling is reproducible. Defaults to ``None``, i.e., no seed.
        evalu (bool, optional): Parameter only needed for development and evaluation purposes. Defaults to ``False``."""

    _report_base: DpMobilityReport
    _report_alternative: DpMobilityReport
    _smape: dict
    _jsd: dict
    _kld: dict
    _emd: dict
    _kt: dict
    _top_n_cov: dict
    _measure_selection: dict
    _top_n_ranking: List[int]
    _similarity_measures: dict = {}

    def __init__(
        self,
        df_base: DataFrame,
        tessellation: Optional[GeoDataFrame] = None,
        df_alternative: Optional[DataFrame] = None,
        privacy_budget_base: Optional[Union[int, float]] = None,
        privacy_budget_alternative: Optional[Union[int, float]] = None,
        user_privacy_base: bool = True,
        user_privacy_alternative: bool = True,
        max_trips_per_user_base: Optional[int] = None,
        max_trips_per_user_alternative: Optional[int] = None,
        analysis_selection: Optional[List[str]] = None,
        analysis_exclusion: Optional[List[str]] = None,
        budget_split_base: dict = {},
        budget_split_alternative: dict = {},
        timewindows: Union[List[int], np.ndarray] = [2, 6, 10, 14, 18, 22],
        max_travel_time: int = 120,
        bin_range_travel_time: int = 5,
        max_jump_length: Union[int, float] = 10,
        bin_range_jump_length: Union[int, float] = 1,
        max_radius_of_gyration: Union[int, float] = 5,
        bin_range_radius_of_gyration: Union[int, float] = 0.5,
        max_user_tile_count: int = 10,
        bin_range_user_tile_count: int = 1,
        max_user_time_delta: Union[int, float] = 48,
        bin_range_user_time_delta: Union[int, float] = 4,
        top_n_ranking: List[int] = [10, 50, 100],
        measure_selection: dict = None,
        disable_progress_bar: bool = False,
        seed_sampling: int = None,
        evalu: bool = False,
    ) -> None:

        self.disable_progress_bar = disable_progress_bar

        self._report_base = DpMobilityReport(
            df_base,
            tessellation,
            privacy_budget_base,
            user_privacy_base,
            max_trips_per_user_base,
            analysis_selection,
            analysis_exclusion,
            budget_split_base,
            timewindows,
            max_travel_time,
            bin_range_travel_time,
            max_jump_length,
            bin_range_jump_length,
            max_radius_of_gyration,
            bin_range_radius_of_gyration,
            max_user_tile_count,
            bin_range_user_tile_count,
            max_user_time_delta,
            bin_range_user_time_delta,
            disable_progress_bar,
            seed_sampling,
            evalu,
        )
        self.report_base.report

        if df_alternative is None:
            df_alternative = df_base

        self._report_alternative = DpMobilityReport(
            df_alternative,
            tessellation,
            privacy_budget_alternative,
            user_privacy_alternative,
            max_trips_per_user_alternative,
            analysis_selection,
            analysis_exclusion,
            budget_split_alternative,
            timewindows,
            max_travel_time,
            bin_range_travel_time,
            max_jump_length,
            bin_range_jump_length,
            max_radius_of_gyration,
            bin_range_radius_of_gyration,
            max_user_tile_count,
            bin_range_user_tile_count,
            max_user_time_delta,
            bin_range_user_time_delta,
            disable_progress_bar,
            seed_sampling,
            evalu,
        )
        self.report_alternative.report

        self.analysis_exclusion = preprocessing.combine_analysis_exclusion(
            self.report_alternative.analysis_exclusion,
            self.report_base.analysis_exclusion,
        )

        (
            self.report_base._report,
            self.report_alternative._report,
        ) = preprocessing.unify_histogram_bins(
            self.report_base.report,
            self.report_alternative.report,
            self.analysis_exclusion,
        )

        if measure_selection is None:
            measure_selection = b_utils.default_measure_selection()
        else:
            measure_selection = preprocessing.validate_measure_selection(
                measure_selection, self.analysis_exclusion
            )
        self._measure_selection = (
            b_utils.remove_excluded_analyses_from_measure_selection(
                measure_selection, self.analysis_exclusion
            )
        )
        self._top_n_ranking = preprocessing.validate_top_n_ranking(top_n_ranking)

        (
            self._smape,
            self._kld,
            self._jsd,
            self._emd,
            self._kt,
            self._top_n_cov,
        ) = compute_similarity_measures(
            self.analysis_exclusion,
            self.report_alternative.report,
            self.report_base.report,
            self.report_base.tessellation,
            self.top_n_ranking,
            self.disable_progress_bar,
        )

    @property
    def report_base(self) -> "DpMobilityReport":
        """The base DpMobilityReport"""
        return self._report_base

    @property
    def report_alternative(self) -> "DpMobilityReport":
        """The alternative DpMobilityReport"""
        return self._report_alternative

    @property
    def similarity_measures(self) -> dict:
        """Similarity measures according to ``measure_selection``."""
        if not self._similarity_measures:
            self._similarity_measures = get_selected_measures(self)

        return self._similarity_measures

    @property
    def measure_selection(self) -> dict:
        """The specified selected similarity measure for each analysis."""
        return self._measure_selection

    @property
    def top_n_ranking(self) -> List[int]:
        """List of 'top n' options that have been provided as user input that are used for compuation of ranking similarity measures."""
        return self._top_n_ranking

    @property
    def smape(self) -> dict:
        """The symmetric (mean absolute) percentage error, based on the relative error, between base and alternative of all selected analyses, where applicable."""
        return self._smape

    @property
    def kld(self) -> dict:
        """The Kullback-Leibler divergence between base and alternative of all selected analyses, where applicable."""
        return self._kld

    @property
    def jsd(self) -> dict:
        """The Jensen-Shannon divergence between base and alternative of all selected analyses, where applicable."""
        return self._jsd

    @property
    def emd(self) -> dict:
        """The earth mover's distance between base and alternative of all selected analyses, where applicable."""
        return self._emd

    @property
    def kt(self) -> dict:
        """The Kendall correlation coefficient of base and alternative of all selected analyses, where applicable."""
        return self._kt

    @property
    def top_n_cov(self) -> dict:
        """Top n coverage of rankings of base and alternative of all selected analyses, where applicable."""
        return self._top_n_cov

    def to_file(
        self,
        output_file: Union[str, Path],
        disable_progress_bar: Optional[bool] = None,
        top_n_flows: int = 100,
    ) -> None:
        """Write the report to a file.
        By default a name is generated.

        Args:
            output_file: The name or the path of the file to store the ``html`` output.
            disable_progress_bar: if ``False``, no progress bar is shown.
            top_n_flows: Determines how many of the top ``n`` origin-destination flows are displayed. Defaults to 100.
        """
        if disable_progress_bar is None:
            disable_progress_bar = self.disable_progress_bar

        if not isinstance(output_file, Path):
            output_file = Path(str(output_file))

        else:
            if output_file.suffix != ".html":
                suffix = output_file.suffix
                output_file = output_file.with_suffix(".html")
                warnings.warn(
                    f"Extension {suffix} not supported. For now we assume .html was intended. "
                    f"To remove this warning, please use .html or .json."
                )

        output_dir = Path(os.path.splitext(output_file)[0])
        filename = Path(os.path.basename(output_file)).stem

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        create_html_assets(output_dir)

        data, temp_map_folder = render_benchmark_html(
            self, filename, top_n_flows, disable_progress_bar
        )

        create_maps_folder(temp_map_folder, output_dir)

        # clean up temp folder
        rmtree(temp_map_folder, ignore_errors=True)

        output_file.write_text(data, encoding="utf-8")



from datetime import timedelta
from typing import TYPE_CHECKING, Optional

import matplotlib.pyplot as plt

if TYPE_CHECKING:
    from dp_mobility_report import DpMobilityReport, BenchmarkReport

from dp_mobility_report import constants as const
from dp_mobility_report.model.section import TupleSection
from dp_mobility_report.report.html.html_utils import (
    all_available_measures,
    get_template,
    render_benchmark_summary,
    render_eps,
    render_moe_info,
    render_summary,
    render_user_input_info,
)
from dp_mobility_report.visualization import plot, v_utils


def render_user_analysis(dpmreport: "DpMobilityReport") -> str:

    args: dict = {}
    report = dpmreport.report

    args[
        "trips_per_user_info"
    ] = f"Trips per user are limited according to the configured maximum of trips per user: {dpmreport.max_trips_per_user}"

    if const.TRIPS_PER_USER not in dpmreport.analysis_exclusion:
        args["trips_per_user_eps"] = render_eps(
            report[const.TRIPS_PER_USER].privacy_budget
        )
        args["trips_per_user_hist"] = render_trips_per_user(
            report[const.TRIPS_PER_USER]
        )
        args["trips_per_user_summary_table"] = render_summary(
            report[const.TRIPS_PER_USER].quartiles
        )
        args["trips_per_user_moe_info"] = render_moe_info(
            report[const.TRIPS_PER_USER].margin_of_error_expmech
        )

    if const.USER_TIME_DELTA not in dpmreport.analysis_exclusion:
        args["time_between_traj_eps"] = render_eps(
            report[const.USER_TIME_DELTA].privacy_budget
        )
        if report[const.USER_TIME_DELTA].quartiles["min"] < timedelta(seconds=0):
            args[
                "plausi_check_info"
            ] = """<strong>Plausibility check</strong>: 
            There are overlapping trips in the dataset.
            The negative minimum time delta implies that there is a trip of a user that starts before the previous one has ended. 
            This might be an indication of an error in the dataset."""
        args["time_between_traj_hist"] = render_time_between_traj(
            report[const.USER_TIME_DELTA]
        )
        args["time_between_traj_summary_table"] = render_summary(
            report[const.USER_TIME_DELTA].quartiles
        )
        args["time_between_traj_moe_info"] = render_moe_info(
            report[const.USER_TIME_DELTA].margin_of_error_expmech
        )

    if const.RADIUS_OF_GYRATION not in dpmreport.analysis_exclusion:
        args["radius_of_gyration_eps"] = render_eps(
            report[const.RADIUS_OF_GYRATION].privacy_budget
        )
        args["radius_of_gyration_hist_info"] = render_user_input_info(
            dpmreport.max_radius_of_gyration, dpmreport.bin_range_radius_of_gyration
        )
        args["radius_of_gyration_hist"] = render_radius_of_gyration(
            report[const.RADIUS_OF_GYRATION]
        )
        args["radius_of_gyration_summary_table"] = render_summary(
            report[const.RADIUS_OF_GYRATION].quartiles
        )
        args["radius_of_gyration_moe_info"] = render_moe_info(
            report[const.RADIUS_OF_GYRATION].margin_of_error_expmech
        )

    if const.USER_TILE_COUNT not in dpmreport.analysis_exclusion:
        args["distinct_tiles_user_eps"] = render_eps(
            report[const.USER_TILE_COUNT].privacy_budget
        )
        args["distinct_tiles_user_hist"] = render_distinct_tiles_user(
            report[const.USER_TILE_COUNT]
        )
        args["distinct_tiles_user_summary_table"] = render_summary(
            report[const.USER_TILE_COUNT].quartiles
        )
        args["distinct_tiles_moe_info"] = render_moe_info(
            report[const.USER_TILE_COUNT].margin_of_error_expmech
        )

    if const.MOBILITY_ENTROPY not in dpmreport.analysis_exclusion:
        args["mobility_entropy_eps"] = render_eps(
            report[const.MOBILITY_ENTROPY].privacy_budget
        )
        args["mobility_entropy_hist"] = render_mobility_entropy(
            report[const.MOBILITY_ENTROPY]
        )
        args["mobility_entropy_summary_table"] = render_summary(
            report[const.MOBILITY_ENTROPY].quartiles
        )
        args["mobility_entropy_moe_info"] = render_moe_info(
            report[const.MOBILITY_ENTROPY].margin_of_error_expmech
        )

    template_structure = get_template("user_analysis_segment.html")

    return template_structure.render(args)


def render_benchmark_user_analysis(benchmark: "BenchmarkReport") -> str:

    args: dict = {}
    report_base = benchmark.report_base.report
    report_alternative = benchmark.report_alternative.report
    template_measures = get_template("similarity_measures.html")

    args[
        "trips_per_user_info"
    ] = f"Trips per user are limited according to the configured maximum of trips per user: {benchmark.report_base.max_trips_per_user} (base); {benchmark.report_alternative.max_trips_per_user} (alternative)."

    if const.TRIPS_PER_USER not in benchmark.analysis_exclusion:
        args["trips_per_user_eps"] = (
            render_eps(report_base[const.TRIPS_PER_USER].privacy_budget),
            render_eps(report_alternative[const.TRIPS_PER_USER].privacy_budget),
        )

        # args["trips_per_user_hist"] = render_trips_per_user(report_base[const.TRIPS_PER_USER],report_alternative[const.TRIPS_PER_USER]) #TODO bins need to be the same
        args["trips_per_user_summary_table"] = render_benchmark_summary(
            report_base[const.TRIPS_PER_USER].quartiles,
            report_alternative[const.TRIPS_PER_USER].quartiles,
            target_type=int,
        )
        args["trips_per_user_moe_info"] = render_moe_info(
            report_base[const.TRIPS_PER_USER].margin_of_error_expmech
        )
        args["trips_per_user_measure"] = template_measures.render(
            all_available_measures(const.TRIPS_PER_USER, benchmark)
        )

        args["trips_per_user_summary_measure"] = template_measures.render(
            all_available_measures(const.TRIPS_PER_USER_QUARTILES, benchmark)
        )

    if const.USER_TIME_DELTA not in benchmark.analysis_exclusion:
        args["time_between_traj_eps"] = (
            render_eps(report_base[const.USER_TIME_DELTA].privacy_budget),
            render_eps(report_alternative[const.USER_TIME_DELTA].privacy_budget),
        )
        args["time_between_traj_hist"] = render_time_between_traj(
            report_base[const.USER_TIME_DELTA],
            report_alternative[const.USER_TIME_DELTA],
        )
        args["time_between_traj_summary_table"] = render_benchmark_summary(
            report_base[const.USER_TIME_DELTA].quartiles,
            report_alternative[const.USER_TIME_DELTA].quartiles,
        )
        args["time_between_traj_moe_info"] = render_moe_info(
            report_base[const.USER_TIME_DELTA].margin_of_error_expmech
        )
        args["time_between_traj_measure"] = template_measures.render(
            all_available_measures(const.USER_TIME_DELTA, benchmark)
        )

        args["time_between_traj_summary_measure"] = template_measures.render(
            all_available_measures(const.USER_TIME_DELTA_QUARTILES, benchmark)
        )

    if const.RADIUS_OF_GYRATION not in benchmark.analysis_exclusion:
        args["radius_of_gyration_eps"] = (
            render_eps(report_base[const.RADIUS_OF_GYRATION].privacy_budget),
            render_eps(report_alternative[const.RADIUS_OF_GYRATION].privacy_budget),
        )

        args["radius_of_gyration_hist_info"] = render_user_input_info(
            benchmark.report_base.max_radius_of_gyration,
            benchmark.report_base.bin_range_radius_of_gyration,
        )
        args["radius_of_gyration_hist"] = render_radius_of_gyration(
            report_base[const.RADIUS_OF_GYRATION],
            report_alternative[const.RADIUS_OF_GYRATION],
        )
        args["radius_of_gyration_summary_table"] = render_benchmark_summary(
            report_base[const.RADIUS_OF_GYRATION].quartiles,
            report_alternative[const.RADIUS_OF_GYRATION].quartiles,
            target_type=float,
        )
        args["radius_of_gyration_moe_info"] = render_moe_info(
            report_base[const.RADIUS_OF_GYRATION].margin_of_error_expmech
        )
        args["radius_of_gyration_measure"] = template_measures.render(
            all_available_measures(const.RADIUS_OF_GYRATION, benchmark)
        )

        args["radius_of_gyration_summary_measure"] = template_measures.render(
            all_available_measures(const.RADIUS_OF_GYRATION_QUARTILES, benchmark)
        )

    if const.USER_TILE_COUNT not in benchmark.analysis_exclusion:
        args["distinct_tiles_user_eps"] = (
            render_eps(report_base[const.USER_TILE_COUNT].privacy_budget),
            render_eps(report_alternative[const.USER_TILE_COUNT].privacy_budget),
        )

        args["distinct_tiles_user_hist"] = render_distinct_tiles_user(
            report_base[const.USER_TILE_COUNT],
            report_alternative[const.USER_TILE_COUNT],
        )
        args["distinct_tiles_user_summary_table"] = render_benchmark_summary(
            report_base[const.USER_TILE_COUNT].quartiles,
            report_alternative[const.USER_TILE_COUNT].quartiles,
            target_type=int,
        )
        args["distinct_tiles_moe_info"] = render_moe_info(
            report_base[const.USER_TILE_COUNT].margin_of_error_expmech
        )
        args["distinct_tiles_measure"] = template_measures.render(
            all_available_measures(const.USER_TILE_COUNT, benchmark)
        )

        args["distinct_tiles_summary_measure"] = template_measures.render(
            all_available_measures(const.USER_TILE_COUNT_QUARTILES, benchmark)
        )

    if const.MOBILITY_ENTROPY not in benchmark.analysis_exclusion:
        args["mobility_entropy_eps"] = (
            render_eps(report_base[const.MOBILITY_ENTROPY].privacy_budget),
            render_eps(report_alternative[const.MOBILITY_ENTROPY].privacy_budget),
        )
        args["mobility_entropy_hist"] = render_mobility_entropy(
            report_base[const.MOBILITY_ENTROPY],
            report_alternative[const.MOBILITY_ENTROPY],
        )
        args["mobility_entropy_summary_table"] = render_benchmark_summary(
            report_base[const.MOBILITY_ENTROPY].quartiles,
            report_alternative[const.MOBILITY_ENTROPY].quartiles,
            target_type=float,
        )
        args["mobility_entropy_moe_info"] = render_moe_info(
            report_base[const.MOBILITY_ENTROPY].margin_of_error_expmech
        )
        args["mobility_entropy_measure"] = template_measures.render(
            all_available_measures(const.MOBILITY_ENTROPY, benchmark)
        )

        args["mobility_entropy_summary_measure"] = template_measures.render(
            all_available_measures(const.MOBILITY_ENTROPY_QUARTILES, benchmark)
        )

    template_structure = get_template("user_analysis_segment_benchmark.html")

    return template_structure.render(args)


def render_trips_per_user(
    trips_per_user_hist: TupleSection,
    trips_per_user_hist_alternative: Optional[TupleSection] = None,
) -> str:
    if trips_per_user_hist_alternative:
        alternative_data = trips_per_user_hist_alternative.data
        alternative_moe = trips_per_user_hist_alternative.margin_of_error_laplace
    else:
        alternative_data = None
        alternative_moe = None
    hist = plot.histogram(
        hist=trips_per_user_hist.data,
        hist_alternative=alternative_data,
        x_axis_label="Number of trips per user",
        y_axis_label="% of users",
        x_axis_type=int,
        margin_of_error=trips_per_user_hist.margin_of_error_laplace,
        margin_of_error_alternative=alternative_moe,
    )
    return v_utils.fig_to_html(hist)


def render_time_between_traj(
    time_between_traj_hist: TupleSection,
    time_between_traj_hist_alternative: Optional[TupleSection] = None,
) -> str:
    if time_between_traj_hist_alternative:
        alternative_data = time_between_traj_hist_alternative.data
        alternative_moe = time_between_traj_hist_alternative.margin_of_error_laplace
    else:
        alternative_data = None
        alternative_moe = None
    hist = plot.histogram(
        hist=time_between_traj_hist.data,
        hist_alternative=alternative_data,
        x_axis_label="Hours between consecutive trips",
        y_axis_label="% of trips",
        x_axis_type=float,
        ndigits_x_label=1,
        margin_of_error=time_between_traj_hist.margin_of_error_laplace,
        margin_of_error_alternative=alternative_moe,
    )
    html = v_utils.fig_to_html(hist)
    plt.close()
    return html


def render_radius_of_gyration(
    radius_of_gyration_hist: TupleSection,
    radius_of_gyration_hist_alternative: Optional[TupleSection] = None,
) -> str:
    if radius_of_gyration_hist_alternative:
        alternative_data = radius_of_gyration_hist_alternative.data
        alternative_error = radius_of_gyration_hist_alternative.margin_of_error_laplace
    else:
        alternative_data = None
        alternative_error = None
    hist = plot.histogram(
        hist=radius_of_gyration_hist.data,
        hist_alternative=alternative_data,
        x_axis_label="radius of gyration (in km)",
        y_axis_label="% of users",
        x_axis_type=float,
        margin_of_error=radius_of_gyration_hist.margin_of_error_laplace,
        margin_of_error_alternative=alternative_error,
    )
    html = v_utils.fig_to_html(hist)
    plt.close()
    return html


def render_distinct_tiles_user(
    user_tile_count_hist: TupleSection,
    user_tile_count_hist_alternative: Optional[TupleSection] = None,
) -> str:
    if user_tile_count_hist_alternative:
        alternative_data = user_tile_count_hist_alternative.data
        alternative_error = user_tile_count_hist_alternative.margin_of_error_laplace
    else:
        alternative_data = None
        alternative_error = None
    hist = plot.histogram(
        hist=user_tile_count_hist.data,
        hist_alternative=alternative_data,
        x_axis_label="number of distinct tiles a user has visited",
        y_axis_label="% of users",
        x_axis_type=int,
        margin_of_error=user_tile_count_hist.margin_of_error_laplace,
        margin_of_error_alternative=alternative_error,
    )
    html = v_utils.fig_to_html(hist)
    plt.close()
    return html


def render_mobility_entropy(
    mobility_entropy: TupleSection,
    mobility_entropy_alternative: Optional[TupleSection] = None,
) -> str:
    if mobility_entropy_alternative:
        alternative_data = (
            mobility_entropy_alternative.data[0],
            mobility_entropy_alternative.data[1].round(2),
        )
        alternative_error = mobility_entropy_alternative.margin_of_error_laplace
    else:
        alternative_data = None
        alternative_error = None
    hist = plot.histogram(
        hist=(mobility_entropy.data[0], mobility_entropy.data[1].round(2)),
        hist_alternative=alternative_data,
        min_value=mobility_entropy.quartiles["min"],
        x_axis_label="mobility entropy",
        y_axis_label="% of users",
        margin_of_error=mobility_entropy.margin_of_error_laplace,
        margin_of_error_alternative=alternative_error,
    )
    html = v_utils.fig_to_html(hist)
    plt.close()
    return html
